/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.backup.plugins;

import de.bsvrz.ars.ars.backup.BackupException;
import de.bsvrz.ars.ars.backup.BackupModul;
import de.bsvrz.ars.ars.backup.MediumNotAccessibleException;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.*;

/**
 * Diese Klasse ist fuer die Sicherung von Containern auf einer DVD verantwortlich. Die Sicherung funktioniert derzeit nur auf Linux-Systemen. Fuer die
 * vorbereitende Erstellung der zu brennenden Verzeichnisstruktur wird das <code>FileSystemBackup</code> verwendet.
 *
 * @author beck et al. projects GmbH
 * @author Christian Wied
 * @version $Revision: 8331 $ / $Date: 2010-11-17 09:35:55 +0100 (Wed, 17 Nov 2010) $ / ($Author: jh $)
 */
public class DVDBackup extends FileSystemBackup implements BackupModul {

	private static final String PROP_BACKUP_DIR = "tempSicherungsVerzeichnis";

	private static final String PROP_BACKUP_SERIALS = "brennerSeriennummern";

	// DVD Standard, wird aus Propertiesfile berschrieben
	private static int MAX_MEDIUM_SIZE_KB = 4300 * 1024;

	private static final String SCSIUSBDirPath = "/proc/scsi/usb-storage";

	protected static final String MEDIAPATH_PREFIX = "/media/usb-";

	private static final String MEDIAPATH_POSTFIX = ":0:0:0";

	private String[] serialNumbers;

	private String[] currentScsiBus;

	private String[] currentSerialID;

	private int EXITCODE_LAST_EXEC = -1; // wird auf den Exit Code eines externen Programms bei exec gesetzt

	public DVDBackup() {
	}

	/** @see de.bsvrz.ars.ars.backup.BackupModul#initialize(java.util.Properties) */
	public void initialize(Properties props) throws Exception {

		if(!amIRunningOnLinux()) throw new Exception("DVD-Backup nicht verfuegbar fuer das Betriebssystem '" + System.getProperty("os.name") + "'.");

		String zipSize = props.getProperty(PROP_BACKUP_CONTAINER_PER_ZIP);
		try {
			if(zipSize != null) {
				MAX_CONTAINERS_PER_DIR = Integer.parseInt(zipSize);
			}
		}
		catch(NumberFormatException e) {
		} // do nothing, default 100 is ok.

		String mediumSize = props.getProperty(PROP_BACKUP_MEDIUMSIZE);
		try {
			if(mediumSize != null) {
				MAX_MEDIUM_SIZE_KB = Integer.parseInt(mediumSize);
			}
		}
		catch(NumberFormatException e) {
		} // do nothing, take default.

		String backupDirName = props.getProperty(PROP_BACKUP_DIR);
		if(backupDirName == null) {
			throw new Exception("Filesystem-Einstellungen sind unvollstndig (" + PROP_BACKUP_DIR + ")");
		}
		File backupDir = new File(backupDirName);
		if(!backupDir.exists()) {
			backupDir.mkdir();
		}
		backupBasePath = backupDir.getAbsolutePath();

		String ser = props.getProperty(PROP_BACKUP_SERIALS);
		if(ser == null || ser.length() == 0) throw new Exception("'brennerSeriennummern' ist nicht angegeben oder enthlt keine Brenner-Seriennummer(n)");
		serialNumbers = ser.split(",");
	}

	public void backupContainer(int mediumID, long contID, File contFile) throws BackupException {
		super.backupContainer(mediumID, contID, contFile);
	}

	/** @see de.bsvrz.ars.ars.backup.BackupModul#closeMediumAfterBackup(int,java.io.File) */
	public void closeMediumAfterBackup(int mediumID, File indexFile) throws BackupException {
		super.closeMediumAfterBackup(mediumID, indexFile);
		String tempPath = backupBasePath + File.separator + "Medium_" + mediumID;

		// checken, ob ueberhaupt Container gesichert werden muessen:
		if(indexFile.length() < 10) { //TODO: mysterise 10
			Util.deleteDir(new File(tempPath));
			logger.fine("Keine Daten zur Sicherung auf DVD vorhanden.");
			return;
		}
		else {
			logger.info("Daten zur Sicherung auf DVD gesammelt.");
		}

		if(!amIRunningOnLinux()) {
			logger.warning("Sicherung auf DVD wird unter " + System.getProperty("os.name") + " nicht unterstuetzt.");
			throw new BackupException("Sicherung auf DVD wird unter " + System.getProperty("os.name") + " nicht unterstuetzt.");
		}

		// Sicherheitshalber nochmals beschreibbare DVD anfordern (ist auch schon in
		// openMediumBeforeBackup() passiert).
		try {
			findEmptyDVD(mediumID);
		}
		catch(MediumNotAccessibleException me) {
			MessageSender.getInstance().sendMessage(
					MessageType.APPLICATION_DOMAIN, MessageGrade.WARNING, "Bitte legen Sie ein neues Medium ein. MediumID=" + mediumID
			);
			logger.warning("Medium neu einlegen");
			throw me;
		}
		catch(Exception e) {
			throw new BackupException(e);
		}
		try {
			// ISO-Image erstellen:
			String command = "mkisofs -udf -T -R -J -quiet -VMedium_" + mediumID + " -o medium" + mediumID + ".iso " + tempPath;
			logger.info("ISO-Image wird erstellt: " + command);
			String output = exec(command);
			if(containsErrorMessage(output) || EXITCODE_LAST_EXEC != 0) throw new BackupException("Fehler bei mkisofs: " + output);

			// ISO-Image auf DVD brennen:
			// Verfgbare Brenner durchpropieren (findEmptyDVD htte eine Ausnahme ausgelst, wenn gar keine DVD
			// verfgbar wre)
			boolean success = false;
			for(int i = 0; i < currentScsiBus.length; i++) {
				if(success) break; // es wurde schon erfolgreich gebrannt
				success = true;
				command = "sudo cdrecord dev=" + currentScsiBus[i] + " gracetime=3 -sao -eject medium" + mediumID + ".iso ";
				logger.info("DVD-Brennen wird gestartet: " + command);
				output = exec(command);
				if(containsErrorMessage(output) || EXITCODE_LAST_EXEC != 0) success = false;

				try {
					Thread.sleep(2000); // soviel Zeit muss sein
				}
				catch(InterruptedException e) {
				} // Nicht raushuepfen, weil sonst der Task beendet wird

				// danach nochmal msinfo aufrufen, damit das Medium vom Laufwerk neu geladen wird.
				// Ansonsten kommt immer "no medium found"
				exec("sudo cdrecord dev=" + currentScsiBus[i] + " -msinfo");
			}
			if(success) {
				logger.info("Brennen auf DVD war erfolgreich.");
			}
			else {
				throw new Exception("cdrecord meldet Fehler: " + EXITCODE_LAST_EXEC);
			}
		}
		catch(Exception e) {
			MessageSender.getInstance().sendMessage(
					MessageType.APPLICATION_DOMAIN, MessageGrade.WARNING, "Bitte legen Sie ein neues Medium ein. MediumID=" + mediumID
			);
			logger.warning("Beim Brennen auf DVD trat folgender Fehler auf: ", e);
			throw new BackupException(e);
		}
		finally {
			// im Gegensatz zum FileSystemBackup wollen wir das temporre Zielverzeichnis
			// nach dem erfolgreichen Schreiben auf DVD natuerlich wieder loeschen.
			Util.deleteDir(new File(tempPath));
			new File("medium" + mediumID + ".iso").delete();
		}
	}

	/** @see de.bsvrz.ars.ars.backup.BackupModul#hasMediumCapacity(int,long,int) */
	public boolean hasMediumCapacity(int mediumID, long containerSize, int mediumIndexSize) throws BackupException {
		// Wir rufen kein super() auf, da wir die hier definierte MAX_MEDIUM_SIZE_KB verwenden
		// wollen (nicht die des
		// FileSystemBackups).
		return MAX_MEDIUM_SIZE_KB * 1024 > currentSpaceOccupiedByZipfiles + currentSpaceOccupiedByContainerFiles + containerSize + mediumIndexSize;
	}

	/** @see de.bsvrz.ars.ars.backup.BackupModul#openMediumBeforeBackup(int,java.lang.String) */
	public void openMediumBeforeBackup(int mediumID, String backupRunID) throws BackupException {
		try {
			findEmptyDVD(mediumID);
		}
		catch(MediumNotAccessibleException me) {
			MessageSender.getInstance().sendMessage(
					MessageType.APPLICATION_DOMAIN, MessageGrade.WARNING, " Bitte legen Sie ein neues Medium ein. MediumID=" + mediumID
			);
			logger.warning("Bitte legen Sie ein neues Medium ein. MediumID=" + mediumID);
			throw me;
		}
		catch(Exception e) {
			throw new BackupException(e);
		}
		logger.info("MediumID=" + mediumID + " / BackupRunId=" + backupRunID + " geoeffnet: "); // +
		// getDVDMediaPath(currentSerialID));
		super.openMediumBeforeBackup(mediumID, backupRunID);
	}

	public InputStream restoreContainer(int mediumID, String contFileName) throws BackupException {
		String tmpDir = backupBasePath;
		try {
			findReadableDVD(mediumID);
		}
		catch(Exception e) {
			logger.warning("Container " + contFileName + " konnte nicht von MediumID=" + mediumID + " wiederhergestellt werden", e);
			throw new BackupException("Container " + contFileName + " konnte nicht von MediumID=" + mediumID + " wiederhergestellt werden");
		}
		for(int i = 0; i < currentSerialID.length; i++) {
			backupBasePath = getDVDMediaPath(currentSerialID[i]);
			try {
				InputStream is = super.restoreContainer(mediumID, contFileName);
				return is;
			}
			catch(Exception e) {
				// do nothing, try in next run
			}
			finally {
				backupBasePath = tmpDir;
			}
		}
		// wenn Container nicht gefunden wurde und nicht oben schon ein InputStream returned wurde:
		throw new BackupException("Container " + contFileName + " konnte nicht von MediumID=" + mediumID + " wiederhergestellt werden");
	}

	/**
	 * Sucht nach leeren Laufwerk (leer = keine DVD oder leere DVD!)
	 *
	 * @param mediumID
	 *
	 * @throws Exception
	 */
	private void findEmptyDVD(int mediumID) throws Exception {
		ArrayList<String> bus = new ArrayList<String>(serialNumbers.length);
		ArrayList<String> id = new ArrayList<String>(serialNumbers.length);
		for(String sn : serialNumbers) {
			if(!containsData(sn)) {
				String tmp = getScsiBus(sn);
				if(tmp != null) {
					bus.add(tmp);
					id.add(sn);
				}
			}
		}
		currentScsiBus = (String[])bus.toArray(new String[0]);
		currentSerialID = (String[])id.toArray(new String[0]);
		if(currentScsiBus.length == 0) {
			logger.info("Keine leere DVD in den Laufwerken gefunden.");
			throw new MediumNotAccessibleException(mediumID);
		}
	}

	private void findReadableDVD(int mediumID) throws Exception {
		ArrayList<String> bus = new ArrayList<String>(serialNumbers.length);
		ArrayList<String> id = new ArrayList<String>(serialNumbers.length);
		for(String sn : serialNumbers) {
			if(isReadable(sn)) {
				String tmp = getScsiBus(sn);
				if(tmp != null) {
					bus.add(tmp);
					id.add(sn);
				}
			}
		}
		currentScsiBus = (String[])bus.toArray(new String[0]);
		currentSerialID = (String[])id.toArray(new String[0]);
		if(currentScsiBus.length == 0) {
			logger.info("DVD mit Medium ID " + mediumID + " nicht gefunden");
			throw new MediumNotAccessibleException(mediumID);
		}
	}

	private boolean isReadable(String serialId) throws Exception {
		String content = getDirContent(serialId);
		if(!containsData(serialId) || !content.contains("Backup_")) {
			logger.finer("DVD in Brenner " + getDVDMediaPath(serialId) + " kann nicht gelesen werden: \n " + content);
			return false;
		}
		logger.finer("DVD in Brenner " + getDVDMediaPath(serialId) + " kann gelesen werden.");
		return true;
	}

	private boolean amIRunningOnLinux() {
		String OS = System.getProperty("os.name").toLowerCase();
		return OS.indexOf("linux") != -1;
	}

	private String getDirContent(String serialId) throws Exception {
		String command = "ls " + getDVDMediaPath(serialId);
		return exec(command);
	}

	/**
	 * Ermittelt, ob es in dem angegebenen Verzeichnis Daten gibt.
	 *
	 * @param serialID
	 *
	 * @return true, wenn es in dem angegebenen Verzeichnis Daten gibt.
	 *
	 * @throws Exception
	 */
	public boolean containsData(String serialID) throws Exception {
		StringTokenizer st = new StringTokenizer(getDirContent(serialID), "\n");
		return st.hasMoreElements();
	}

	/**
	 * Da der SCSI-Bus eines USB-Devices vom Betriebssystem dynamisch vergeben wird (d.h. sich beim Ein-/Ausschalten des Gerts aendern kann), muss die Beziehung
	 * Hardware <---> SCSI-Bus regelmaessig nachgeschlagen werden. Diese Methode bietet die hierfuer benoetigte Funktionalitaet.
	 *
	 * @param serialID
	 *
	 * @return den SCSI-BUS als String mit Format: x,0,0
	 *
	 * @throws IOException
	 */
	String getScsiBus(String serialID) throws IOException {

		/**
		 * Das Verzeichnis enthaelt zwei Dateien mit folgendem Inhalt:<br>
		 * <br>
		 * ls /proc/scsi/usb-storage<br>
		 * <br> . .. 1 3<br>
		 * <br>
		 * cat /proc/scsi/usb-storage/1<br>
		 * Host scsi1: usb-storage<br>
		 * Vendor: HL-DT-ST DVDRAM GSA-5163D<br>
		 * Product: Super Multi DVD<br>
		 * Rewriter Serial Number: 4B13E055<br>
		 * Protocol: Transparent SCSI<br>
		 * Transport: Bulk<br>
		 * Quirks:<br>
		 */
		File usbDir = new File(SCSIUSBDirPath);
		if(!usbDir.exists() || !usbDir.isDirectory()) {
			return null;
		}
		for(File usbDev : usbDir.listFiles()) {
			String output = exec("cat " + usbDev.getAbsolutePath());
			if(output.indexOf(serialID) != -1) {
				return usbDev.getName() + ",0,0";
			}
		}
		return null;
	}

	String getDVDMediaPath(String serialId) {
		return MEDIAPATH_PREFIX + serialId + MEDIAPATH_POSTFIX;
	}

	/**
	 * Fuehrt ein Kommando aus und liefert den Output als String zurueck.<br> Der Output ist auf 100000 Zeichen limitiert.
	 *
	 * @param command
	 *
	 * @return Kommando-Ausgabe
	 *
	 * @throws IOException
	 */
	private String exec(String command) throws IOException {
		logger.fine("Fhre Linuxbefehl aus: " + command);

		Process proc = Runtime.getRuntime().exec(command);
		BufferedReader procStdout = new BufferedReader(new InputStreamReader(proc.getInputStream()));
		String bufStr;
		StringBuffer buf = new StringBuffer();
		while((bufStr = procStdout.readLine()) != null) buf.append(bufStr + "\n");
		procStdout.close();

		try {
			EXITCODE_LAST_EXEC = proc.waitFor();
		}
		catch(InterruptedException e) {
			logger.error("Fehler beim Ausfhren von '" + command + "': ", e);
		}
		logger.finer("stdout des Befehls:" + buf.toString());
		return buf.toString();
	}

	/**
	 * Prft, ob in input das Wort 'Fehler' oder 'Error' vorkommt, unabhngig von Gro/Kleinschreibung.
	 *
	 * @param input Input String
	 *
	 * @return Ob eine Fehlermeldung enthalten ist.
	 */
	private boolean containsErrorMessage(String input) {
		return (input.toLowerCase().contains("error") || input.toLowerCase().contains("fehler"));
	}

	public String toString() {
		return "DVD (temporres Verzeichnis=" + currentBackupPath + ")";
	}
}
