/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DIdNodeNotFoundException;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.tasks.tasklistener.RingOfDeathListener;
import de.bsvrz.ars.ars.mgmt.tasks.tasklistener.TaskStepListener;
import de.bsvrz.ars.ars.persistence.ContainerDataIterator;
import de.bsvrz.ars.ars.persistence.ContainerFile;
import de.bsvrz.ars.ars.persistence.ContainerHdr;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.index.IndexResult;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.ContainerSettings;
import de.bsvrz.sys.funclib.losb.datk.ContainerSettings.CloseCondition;
import de.bsvrz.sys.funclib.losb.ringbuffer.ArchiveRingBuffer;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;

import java.io.File;
import java.util.concurrent.atomic.AtomicLong;
import java.util.zip.Deflater;

/**
 * Task zur Archivierung von Datensaetzen. Ist nicht von SingleTask abgeleitet, weil ein spezieller ArchiveRingBuffer verwendet wird, der zwei Objekte
 * (ResultData und Archivzeit) speichern kann. Die work()-Methode hat deswegen zwei Parameter.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 8179 $ / $Date: 2010-09-20 13:07:40 +0200 (Mon, 20 Sep 2010) $ / ($Author: rs $)
 */
public class ArchiveTask extends Task {

	protected static final boolean REQUESTED = true, ONLINE = false;

	private static final ArchiveDataKind OA = ArchiveDataKind.ONLINE;

	private static final ArchiveDataKind ON = ArchiveDataKind.ONLINE_DELAYED;

	protected static final ArchiveDataKind NA = ArchiveDataKind.REQUESTED;

	protected static final ArchiveDataKind NN = ArchiveDataKind.REQUESTED_DELAYED;

	/** Kennzeichen der Betriebsmeldung */
	public static final String MSG_PID_DATAINDEX_ANOMALY = "Datenindex nicht monoton";

	private static AtomicLong TOTAL_DS_QUEUED = new AtomicLong(0);

	private static AtomicLong TOTAL_DS_SUCCESS = new AtomicLong(0);

	private static AtomicLong TOTAL_DS_FAIL = new AtomicLong(0);

	private static AtomicLong CLOSE_CONT_SUCCESS = new AtomicLong(0);

	protected byte[] idxBuffer = new byte[INDEX_MEMORY_SIZE];

	private ArchiveRingBuffer inputDataQueue;

	private ArchiveRingBuffer.PopObject popObject;	// zum uebertragen des Paares ResultData/Archivzeit

	/** Letzte verwendete Archivzeit. Wird fuer die Archivzeitueberwachung benutzt. */
	private static long lastArchiveTime = 0;

	// Kennzeichnungen des zu archivierenden Datensatzes:
	private ResultData rd;

	protected long objID, aspID, arsTime, dataTime, dataIdx, openContID;

	protected AttributeGroup atg;

	protected int sv;	// ist beim Nachfordern immer 0

	protected ArchiveDataKind adk;

	protected String didPath, didPathA, didPathN;

	protected DataIdentNode din;

	/** Laenge des serislisierten Datensatzes inklusive Datensatz-Header */
	protected int dataEntryLen;

	private long lastADataIdx, lastNDataIdx, lastDataIdx, lastDataTime;

	private long lastAContID;

	private boolean lastDataWasGapMarker;

	protected Deflater compresser = new Deflater();

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// ContainerSettings:
	//
	/** Container-Einstellungen des Archivsystems (aus atg.archivContainer). */
	private static ContainerSettings containerSettings;

	static {
		// Standardwerte (keine Default-Werte im Datenkatalog vorgesehen):
		containerSettings = new ContainerSettings();
		containerSettings.stdCloseConditions.maxContAnzDS = 5000;
		containerSettings.stdCloseConditions.maxContSize = 20 * 1000 * 1000;	// 20 MB
		containerSettings.stdCloseConditions.maxContTime = 60 * 60 * 24 * 7;	// 1 Woche
	}

	/**
	 * Setzt die Container-Einstellungen des Archivsystems.
	 *
	 * @param cs Container-Einstellungen
	 */
	public static void setContainerSettings(ContainerSettings cs) {
		containerSettings = cs;
	}

	/**
	 * Liefert die Container-Einstellungen des Archivsystems.
	 *
	 * @return Container-Einstellungen
	 */
	public static ContainerSettings getContainerSettings() {
		return containerSettings;
	}
	//
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Konstruktor des Archiv-Tasks.
	 *
	 * @param tMgr Task-Manager
	 * @param pMgr Persistenz-Manager
	 */
	public ArchiveTask(ArchiveManager archiveMgr, RingOfDeathListener rodl, int tskIdx, ArchiveRingBuffer inQueue) {
		super(archiveMgr, rodl);
		logger = Debug.getLogger();
		setName(getName() + "_" + tskIdx);
		inputDataQueue = inQueue;
		popObject = inQueue.new PopObject();	// Zum uebergeben von Paaren (Data, Archivzeit)
	}

	/* (non-Javadoc)
		 * @see java.lang.Runnable#run()
		 */
	public void run() {
		try {
			while(!isTaskTerminated()) {
				try {
					inputDataQueue.pop(popObject);		// synchronized, blockiert falls leer
					archMgr.getInQueuesMgr().decrOnlineDataInQueues();
					work(popObject.object, popObject.time);

					synchronized(taskStepListeners) {
						for(TaskStepListener tsl : taskStepListeners) {
							tsl.taskStepDone(this);
						}
					}
				}
				catch(InterruptedException e) {
					break;
				}
				catch(Exception e) {
					logger.error("Schwerer Fehler", e);
				}
			}
		}
		finally {
			if(ringOfDeathListener != null) ringOfDeathListener.ringOfDeath(this);
		}
	}

	/**
	 * Wird vom {@link de.bsvrz.ars.ars.mgmt.InQueuesMgr.ArchiveDataReceiver} aufgerufen, nachdem er einen zu archivierenden Datensatz in eine der Queues gelegt
	 * hat.
	 */
	public static void incrementTotalDS_Queued() {
		TOTAL_DS_QUEUED.incrementAndGet();
	}

	/** @return Anzahl der seit Systemstart in die Warteschlangen eingefuegten zu archivierenden Datensatze */
	public static long getTotalDS_Queued() {
		return TOTAL_DS_QUEUED.longValue();
	}

	/** @return Anzahl der seit Systemstart zur Archivierung verarbeiteten Datesaetze (erfolgreich und nicht erfolgreich archiviert) */
	public static long getTotalDS_Processed() {
		return TOTAL_DS_SUCCESS.longValue() + TOTAL_DS_FAIL.longValue();
	}

	/** @return Anzahl der seit Systemstart erfolgreich archivierten Datesaetze */
	public static long getTotalDS_Success() {
		return TOTAL_DS_SUCCESS.longValue();
	}

	/** @return Anzahl der seit Systemstart nicht erfolgreich archivierten Datesaetze */
	public static long getTotalDS_Fail() {
		return TOTAL_DS_FAIL.longValue();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} erfolgreich verarbeiteten Close-Container-Datensaetze. Kann zur
	 *         Test-Synchronisation verwendet werden.
	 */
	public static long getCloseContainerSuccess() {
		return CLOSE_CONT_SUCCESS.longValue();
	}


	/** Setzt alle Zaehler auf 0. */
	public static void resetDSCounter() {
		TOTAL_DS_SUCCESS.set(0);
		TOTAL_DS_FAIL.set(0);
		TOTAL_DS_QUEUED.set(0);
		CLOSE_CONT_SUCCESS.set(0);
	}

	public static void setLastArchiveTime(long atime) {
		lastArchiveTime = atime;
	}

	public static long getLastArchiveTime() {
		return lastArchiveTime;
	}


	protected void assignIndexes(boolean requested) throws DIdNodeNotFoundException, IndexException, PersistenceException {
		// Es muessen stets beide Indexe (aktuell/nachgeliefert) angelegt werden, da Rueckspruenge 
		// und Luecken ueber beide Datensatzarten gehen.
		if(!din.indexesAssigned(requested ? NA : OA)) persMgr.assignIndexes(didTree, objID, atg.getId(), aspID, sv, (requested ? NA : OA), this, idxBuffer);
		if(!din.indexesAssigned(requested ? NN : ON)) persMgr.assignIndexes(didTree, objID, atg.getId(), aspID, sv, (requested ? NN : ON), this, idxBuffer);
	}

	/**
	 * Archiviert den uebergebenen Datensatz unter dem angegebenen Archivzeitstempel.
	 *
	 * @param resultData Ergebnisdatensatz
	 * @param archTime   Archivzeitstempel
	 */
	public void work(Object resultData, long archTime) {
		// Ist der DS ein vom ArS erzeugter DS (z.B. um Cont.-Abschluss zu erzwingen)?
		// archTime ist dann in Wahrheit die Object-ID
		if((resultData instanceof CloseContainerObject)) {
			closeContainer(archTime, (CloseContainerObject)resultData); // Sync'ed selbst auf DIN
			return;
		}

		logReceivedData((ResultData)resultData);

		// Synchronisierung ueber den DataIdentNode:
		boolean criticalAreaEntered = false;
		try {
			presetData((ResultData)resultData, archTime);

			criticalAreaEntered = din.enterCriticalArea();
			if(!criticalAreaEntered) {
				logger.warning("Archivierung abgebrochen: " + resultData);
				return;
			}

			assignIndexes(ONLINE);			// Indexe zuweisen, falls nicht schon geschehen
			lookupLastDataIdxes(ONLINE);	// Letzte Datenindexe, Datenzeiten nachsehen	
			if(handleIndexBackstep()) return;

			openContID = din.getOpenContID(adk);

			// Luecke? (Wenn der letze OA-Datensatz "keine Quelle" war, keine Luecke einfuegen)
			// Nicht extra pruefen, wenn beim letzten Aufruf PotGap eingefuegt wurde und diesmal der echte DS.
			// Aus Performance-Gruenden wird im DIN vermerkt, ob der letzte DS ein "keine Quelle" war. Damit
			// spart man den Dateizugriff.
			boolean insertGapDS = false;
			insertGapDS = !lastDataWasGapMarker && isDataGap() && (din.lastDataNoSourceAssigned() ? !din.lastDataWasNoSource() : !scanOAContForLastNoSrc());
			if(insertGapDS) {
				logDataGap();
				// Luecken werden eingefuegt, indem zuerst der Datensatz auf einen Luecken-Datensatz umgesetzt
				// wird und danach work() mit dem urspr. Datensatz nochmal aufgerufen wird. Alles andere bleibt gleich.
				// Luecken werden immer im OA-Cont. eingetragen. Also ggf. Datensatzart wechseln.					
				if(!adk.equals(OA)) {
					adk = OA;
					didPath = didPathA;
					openContID = din.getOpenContID(OA);
				}
				// DIdx, Datenzeit der Luecke gemaess TAnfArS 5.1.2.4.3.5 setzen
				dataIdx = Util.dIdxSetArSBit(lastDataIdx);
				dataTime = lastDataTime;
				dataEntryLen = persMgr.serializeDataGap(this, dataIdx, dataTime, arsTime, compresser);
			}
			else {
				dataEntryLen = persMgr.serializeData(this, rd, arsTime, compresser);
			}

			if(storeData()) {
				din.setFirstDataAfterSubscription(adk, false);							// ab jetzt duerfen Datenindexe nicht mehr gleich sein
				din.setLastOAWasNoSource(
						adk.equals(OA) && rd.isNoSourceAvailable()
				);	// Fuer naechsten DS merken, damit man nicht in den Cont. schauen muss
				logWriteData();
				if(!insertGapDS) {
					TOTAL_DS_SUCCESS.incrementAndGet();	// Datenlueckensaetze zaehlen nicht
					if(rd.hasData()) sendAck(rd);	// nicht quittieren, falls kQ/kD oder nur eine Datenlcke eingefuegt wurde
				}
			}
			else {
				TOTAL_DS_FAIL.incrementAndGet();
			}

			// Vermerken, dass als naechstes der echte DS ausser der Reihe mit work() eingefuegt wird.
			// Da dieser ebenfalls eine Luecke zum Gap-DS hat, muss verhindert werden, dass wieder eine
			// Datenluecke festgestellt wird.
			lastDataWasGapMarker = insertGapDS;
			if(lastDataWasGapMarker) {
				// Letzter DS war Datenluecke: Einfuegen des echten Datensatzes nachholen.
				// Dazu alles so verlassen, dass die Methode wieder normal ablaufen kann:
				persMgr.leaveContainer(this);
				din.exitCriticalArea();
				work(rd, arsTime);
			}
		}
		catch(DIdNodeNotFoundException e) {
			handleErr(e, Util.rd2Str(rd));
		}	// Darf nicht sein, weil die DID zum Archivieren parametriert wurde
		catch(PersistenceException e) {
			handleErr(e, Util.rd2Str(rd));
		}
		catch(IndexException e) {
			handleErr(e, Util.rd2Str(rd));
		}
		catch(Throwable e) {
			handleSevereErr(e, Util.rd2Str(rd));
		}

		finally {
			if(din != null && criticalAreaEntered) din.exitCriticalArea();
		}
	}

	protected boolean storeData() throws PersistenceException, IndexException {
		try {
			if(openContID == -1) {
				// Container entweder abgeschlossen oder es gibt keinen, also neuen anlegen				
				if(tooLargeForNewContainer()) return false;
				openContID = persMgr.nextContainerID();
				persMgr.accessContainer(this, objID, atg.getId(), aspID, sv, adk, didPath, true, openContID);
				persMgr.createContainer(this, din.arSParamIsSichern());
				logNewContainer();
				persMgr.appendSerializedData(this);
				din.addIndexEntries(adk, didPath, dataIdx, arsTime, dataTime, openContID, idxBuffer);
			}
			else {
				persMgr.accessContainer(this, objID, atg.getId(), aspID, sv, adk, didPath, false, openContID);
				if(mustCloseContainer(false)) {
					persMgr.closeContainer(this, din, adk);
					persMgr.leaveContainer(this);
					if(tooLargeForNewContainer()) return false;
					openContID = persMgr.nextContainerID();
					persMgr.accessContainer(this, objID, atg.getId(), aspID, sv, adk, didPath, true, openContID);	// legt notfalls Verzeichnis an
					persMgr.createContainer(this, din.arSParamIsSichern());
					logNewContainer();
					persMgr.appendSerializedData(this);
					din.addIndexEntries(adk, didPath, dataIdx, arsTime, dataTime, openContID, idxBuffer);
				}
				else {
					persMgr.appendSerializedData(this);
					din.updateIndexEntries(adk, dataIdx, arsTime, dataTime);
				}
			}
			createRebuildIndexFlagFile();		// Markieren, dass etwas veraendert wurde
			return true;
		}
		finally {
			persMgr.leaveContainer(this);
		}
	}

	/**
	 * Setzt Flags, IDs, holt den DataIdentNode, usw.
	 *
	 * @param resultData Aktueller Ergebnisdatensatz
	 * @param archTime   Archivzeitstempel
	 *
	 * @throws DIdNodeNotFoundException
	 */
	private void presetData(ResultData resultData, long archTime) throws DIdNodeNotFoundException {
		rd = resultData;
		lastADataIdx = lastNDataIdx = lastDataIdx = lastDataTime = lastAContID = -1;
		DataDescription dd = rd.getDataDescription();
		objID = rd.getObject().getId();
		atg = dd.getAttributeGroup();
		aspID = dd.getAspect().getId();
		sv = dd.getSimulationVariant();
		arsTime = archTime;
		dataTime = rd.getDataTime();
		dataIdx = rd.getDataIndex();
		adk = rd.isDelayedData() ? ON : OA;
		didPathA = persMgr.getPathName(objID, atg.getId(), aspID, sv, Util.OA);
		didPathN = persMgr.getPathName(objID, atg.getId(), aspID, sv, Util.ON);
		didPath = rd.isDelayedData() ? didPathN : didPathA;
		din = didTree.getDataIdentNode(objID, atg.getId(), aspID, sv);
	}

	/**
	 * Schliesst einen Container ab. Wird durch ein spezielles Objekt ausgeloest, das in die Queue der ArchiveTask gelegt wird. Um Platz zu sparen, wird anstelle
	 * der Archivzeit die Objekt-ID gespeichert. Der Aufrufer dieser Methode muss die Synchronisierung ueber den DataIdentNode durchfuehren.
	 *
	 * @param closeObjID
	 * @param cco
	 */
	private void closeContainer(long closeObjID, CloseContainerObject cco) {
		// DataIdentNode besorgen:
		try {
			din = didTree.getDataIdentNode(closeObjID, cco.atgID, cco.aspID, cco.sv);
		}
		catch(DIdNodeNotFoundException e) {
			handleErr(e, cco);
			return;
		}

		// Synchronisierung ueber den DataIdentNode:
		boolean criticalAreaEntered = false;
		try {
			criticalAreaEntered = din.enterCriticalArea();
			if(!criticalAreaEntered) {
				logger.warning("Archivierung abgebrochen");
				return;
			}

			for(int archiveDataKindIndex = Util.OA; archiveDataKindIndex <= Util.NN; archiveDataKindIndex++) {
				adk = Util.getDataKindFromIndex(archiveDataKindIndex);

				if(!din.indexesAssigned(adk)) {
					// Deswegen keine Indexe anlegen
					String path = persMgr.getPathName(closeObjID, cco.atgID, cco.aspID, cco.sv, archiveDataKindIndex);
					openContID = persMgr.getLargestContainerID(path);
					if(openContID == -1) {
						continue;       // Keine Container, also auch keine abzuschliessen
					}
					else {
						try {
							persMgr.accessContainer(this, closeObjID, cco.atgID, cco.aspID, cco.sv, adk, openContID);
							if(persMgr.isContainerClosed(this)) {
								continue;   // Bereits abgeschlossen
							}
							else {
								persMgr.closeContainerRecount(this, din, adk);   // der teuerste Fall...
								logManualCloseContainer(closeObjID, cco);
							}
						}
						catch(PersistenceException e) {
							handleErr(e, cco);
						}
						finally {
							persMgr.leaveContainer(this);
						}
					}
				}
				else {
					openContID = din.getOpenContID(adk);
					// Es kann sein, dass Indexe fuer eine Anfrage angelegt wurden, ohne dass Daten da sind.
					// Drum hier nochmal prfen
					if(openContID != -1) {
						try {
							persMgr.accessContainer(this, closeObjID, cco.atgID, cco.aspID, cco.sv, adk, openContID);
							persMgr.closeContainer(this, din, adk);
							logManualCloseContainer(closeObjID, cco);
						}
						catch(PersistenceException e) {
							handleErr(e, cco);
						}
						finally {
							persMgr.leaveContainer(this);
						}
					}
				}
			}
			CLOSE_CONT_SUCCESS.incrementAndGet();
		}
		catch(Exception e) {
			handleErr(e, cco);
		}
		finally {
			if(din != null && criticalAreaEntered) din.exitCriticalArea();
		}
	}

	/**
	 * Durchlaeuft den aktuellen Container und prueft, ob der letzte Datensatz die Kennung "keine Quelle" hat (langsam!). Dies ist nur notwendig, wenn das erste
	 * mal auf die DID zugegriffen wird. Sonst genuegt {@link DataIdentNode#lastOADataWasNoSource}. Der Aufrufer dieser Methode muss die Synchronisierung ueber den
	 * DataIdentNode durchfuehren.
	 *
	 * @return Wahr, falls der letzte Datensatz die Kennung "keine Quelle" hat, falsch sonst
	 *
	 * @throws PersistenceException
	 */
	private boolean scanOAContForLastNoSrc() throws PersistenceException {
		boolean result = false;
		try {
			if(lastAContID == -1) return false;
			persMgr.accessContainerReadOnly(
					this, new File(didPathA, ContainerFile.getContainerFileName(lastAContID))
			);

			persMgr.readContainerHeader(this);
			final boolean markedAsDeleted = persMgr.getContainerHeaderParamAsBoolean(this, ContainerHdr.CHP_DELETED);
			if(markedAsDeleted) {
				// Wenn der Container als gelscht markiert ist, dann nehmen wir an, dass der letzte Datensatz nicht die Kennung "keine Quelle" hat.
				result = false;
			}
			else {
				// Wenn der Container als nicht als gelscht markiert ist, dann schauen wir uns jeden Datensatz im Container an.
				ContainerDataIterator cdi = persMgr.getArchiveData(this);
				while(cdi.readNextDataHdr()) {
					result = cdi.isNoSource();
					cdi.skipData();
				}
				cdi.close();
			}
		}
		catch(Exception e) {
			
			logger.warning("Fehler beim Scan ber den letzten Container wird ignoriert", e);
			result = false;
		}
		finally {
			persMgr.leaveContainer(this);
		}
		return result;
	}

	/**
	 * Sucht die letzten Datenindexe im Index sowie den Container aus dem . Der Aufrufer dieser Methode muss die Synchronisierung ueber den DataIdentNode
	 * durchfuehren.
	 *
	 * @throws IndexException
	 */
	protected void lookupLastDataIdxes(boolean requested) throws IndexException {
		ArchiveDataKind a = requested ? NA : OA;
		ArchiveDataKind n = requested ? NN : ON;
		lastADataIdx = din.maxDIdx(a);
		lastNDataIdx = din.maxDIdx(n);
		lastAContID = din.getOpenContID(a);
		long lastADTime = din.maxDTime(a);
		long lastNDTime = din.maxDTime(n);
		if(lastADataIdx == -1 && lastNDataIdx == -1) {
			IndexResult irA = din.getDataIdxContIDs(a, didPathA, idxBuffer, 0, Long.MAX_VALUE);
			IndexResult irN = din.getDataIdxContIDs(n, didPathN, idxBuffer, 0, Long.MAX_VALUE);
			lastADataIdx = irA.size() > 0 ? irA.getDIMax(irA.size() - 1) : -1;
			lastNDataIdx = irN.size() > 0 ? irN.getDIMax(irN.size() - 1) : -1;
			lastADTime = irA.size() > 0 ? irA.getDTMax(irA.size() - 1) : -1;
			lastNDTime = irN.size() > 0 ? irN.getDTMax(irN.size() - 1) : -1;
			lastAContID = irA.size() > 0 ? irA.getCID(irA.size() - 1) : -1;
		}
		lastDataIdx = lastADataIdx >= lastNDataIdx ? lastADataIdx : lastNDataIdx;
		lastDataTime = lastADTime > lastNDTime ? lastADTime : lastNDTime;
	}

	/**
	 * Prueft, ob eine potentielle Datenluecke vorliegt. Der Aufrufer dieser Methode muss die Synchronisierung ueber den DataIdentNode durchfuehren.
	 *
	 * @return Wahr, falls eine potentielle Datenluecke vorliegt, falsch sonst.
	 *
	 * @throws IndexException
	 */
	private boolean isDataGap() throws IndexException {
		if(ArchiveManager.ONLY_DATA) return false;	// nur fuer Testzwecke
		if(lastDataIdx == -1) {
			return false;	// noch nie vorher was archiviert
		}

		// "63/93-Fehler": Wenn der letzte DS ein Gap-Marker ist, der naechste DS aber nicht geschrieben
		// wurde, wuerde das Setzen des ArSBits von lastDataIdx nichts aendern und einen Indexfehler 
		// verursachen. Der neue DS wird nicht geschrieben und das Problem wiederholt sich.
		// Daher kann es durch die folgende Abfrage nie eine Luecke geben, wenn der letzte
		// DS eine Lueckenmarkierung ist.
		else if(Util.dIdxArSBit(lastDataIdx) == 1) {
			return false;
		}

		else if(Util.dIdxNoModBits(dataIdx) > Util.dIdxNoModBits(lastDataIdx) + 1) {
			logger.fine(
					"Potentielle Datenluecke erkannt: letzer archivierter Index (O" + (lastADataIdx > lastNDataIdx ? "A" : "N") + "): " + lastDataIdx + " ["
					+ Util.dIdx2Str(lastDataIdx) + "] [" + Util.dIdx2StrExt(lastDataIdx) + "]" + ", erhaltener Datenindex" + dataIdx + " ["
					+ Util.dIdx2Str(dataIdx) + "] [" + Util.dIdx2StrExt(dataIdx) + "]"
			);
			return true;
		}
		else {
			return false;
		}
	}

	/**
	 * Prueft, ob der Datenindex plausibel ist und sendet ggf. die entsprechenden Betriebsmeldungen. Prueft, ob der Datenindex kleiner, gleich oder groesser als
	 * der zuletzt archivierte DI ist. Wenn der DI kleiner ist, wird eine BetrMeld verschickt und nichts archiviert. Wenn der DI gleich ist und dies der erste DS
	 * nach einer Neuanmeldung ist, wird der DS ignoriert (weil der DAV moeglicherweise einen gepufferten DS geschickt hat). Wenn der DI groesser ist, wird normal
	 * archiviert.
	 *
	 * @return Wahr, falls der Index nicht plausibel ist und der Datensatz nicht archiviert werden darf, falsch sonst.
	 *
	 * @throws IndexException
	 */
	private boolean handleIndexBackstep() throws IndexException {
		if(dataIdx == lastDataIdx && din.isFirstDataAfterSubscription(adk)) {
			logger.fine("Datenindex nach Neuanmeldung identisch zum letzten archivierten Datenindex. Datensatz ignoriert." + Debug.NEWLINE, Util.rd2Str(rd));
			return true;
		}
		if(dataIdx <= lastDataIdx) {
			DataDescription dd = rd.getDataDescription();
			StringBuilder sb = new StringBuilder(8192);
			sb.append("Es ist ein fehlerhafter Datenindex aufgetreten. Der soeben erhaltene Datenindex ist nicht streng monoton steigend.");
			sb.append("\n\nSoeben erhalten\n===============");
			sb.append("\nObjekt:              ");
			sb.append(rd.getObject());
			sb.append(" [ID=" + rd.getObject().getId() + "])");
			sb.append("\nAttributgruppe:      ");
			sb.append(dd.getAttributeGroup());
			sb.append(" [ID=" + dd.getAttributeGroup().getId() + "]");
			sb.append("\nAspekt:              ");
			sb.append(dd.getAspect());
			sb.append(" [ID=" + dd.getAspect().getId() + "]");
			sb.append("\nSimVar:              ");
			sb.append(dd.getSimulationVariant());
			sb.append("\nDatenindex:          ");
			sb.append(dataIdx);
			sb.append(" [" + Util.dIdx2Str(dataIdx) + "] [" + Util.dIdx2StrExt(dataIdx) + "]");
			sb.append("\nDatenzeit:           ");
			sb.append(Util.timestrMillisFormatted(rd.getDataTime()));
			sb.append("\nerhalten:            ");
			sb.append(Util.timestrMillisFormatted(arsTime));
			sb.append("\nArt:                 ");
			sb.append(rd.getDataKind());
			sb.append("\nkeine Quelle:        ");
			sb.append(Util.bTF(rd.isNoSourceAvailable()));
			sb.append("\nQuelle verfgbar:    ");
			sb.append(Util.bTF(rd.isSourceAvailable()));
			sb.append("\nkeine Daten:         ");
			sb.append(Util.bTF(rd.isNoDataAvailable()));
			sb.append("\nkeine Rechte:        ");
			sb.append(Util.bTF(rd.isNoRightsAvailable()));
			sb.append("\nungltige Anmeldung: ");
			sb.append(Util.bTF(rd.isNoValidSubscription()));
			sb.append("\n\n**Beginn Daten**\n");
			sb.append(rd);
			sb.append("\n**Ende Daten**\n");
			sb.append("\nZuletzt archiviert\n==================");
			sb.append("\nArt:                 ");
			sb.append(OA);
			if(lastADataIdx < 0) {
				sb.append("\nDatenindex:          -");
				sb.append("\nDatenzeit:           -");
				sb.append("\nArchivzeit:          -");
			}
			else {
				sb.append("\nDatenindex:          ");
				sb.append(lastADataIdx);
				sb.append(" [" + Util.dIdx2Str(lastADataIdx) + "]");
				sb.append(" [" + Util.dIdx2StrExt(lastADataIdx) + "]");
				sb.append("\nDatenzeit:           " + Util.timestrMillisFormatted(din.maxDTime(OA)));
				sb.append("\nArchivzeit:          " + Util.timestrMillisFormatted(din.maxATime(OA)));
			}
			sb.append("\nArt:                 ");
			sb.append(ON);
			if(lastNDataIdx < 0) {
				sb.append("\nDatenindex:          -");
				sb.append("\nDatenzeit:           -");
				sb.append("\nArchivzeit:          -");
			}
			else {
				sb.append("\nDatenindex:          ");
				sb.append(lastNDataIdx);
				sb.append(" [" + Util.dIdx2Str(lastNDataIdx) + "]");
				sb.append(" [" + Util.dIdx2StrExt(lastNDataIdx) + "]");
				sb.append("\nDatenzeit:           " + Util.timestrMillisFormatted(din.maxDTime(ON)));
				sb.append("\nArchivzeit:          " + Util.timestrMillisFormatted(din.maxATime(ON)));
			}
			sb.append("\n");
			logger.warning("Betriebsmeldung: Datenindex nicht monoton", sb.toString());
			MessageSender.getInstance().sendMessage(
					MSG_PID_DATAINDEX_ANOMALY, MessageType.APPLICATION_DOMAIN, "", MessageGrade.FATAL, MessageState.MESSAGE, sb.toString()
			);
			return true;
		}
		return false;
	}

	/**
	 * Sendet eine Quittung fuer den angegebenen Datensatz.
	 *
	 * @param rd Datensatz
	 */
	private void sendAck(ResultData rd) {
		// Der Quittierungsaspekt wird von ArchivConfig sofort abgemeldet, sobald eine DID nicht mehr 
		// archiviert/parametriert/quittiert wird.
		// Daher kann fuer solche bereits in der Queue vorhandenen DS keine Quittung gesendet werden.

		if(din.isArSParameterized() && din.arSParamIsArchivieren() && din.arSParamIsQuittierenValid()) {
			long qAspId = din.arSParamGetQuittieren();
			try {
				archMgr.getInQueuesMgr().getDataAckSender().sendAck(rd, qAspId);
				logger.finest("Quittung unter Aspekt '" + qAspId + "' gesendet fuer", Util.rd2Str(rd));
			}
			catch(DataNotSubscribedException e) {
				logger.warning("Senden der Quittung unter Aspekt '" + qAspId + "' fehlgeschlagen: " + e.getMessage() + Debug.NEWLINE + Util.rd2Str(rd));
			}
			catch(ConfigurationException e) {
				logger.warning("Senden der Quittung unter Aspekt '" + qAspId + "' fehlgeschlagen: " + e.getMessage() + Debug.NEWLINE + Util.rd2Str(rd));
			}
			catch(SendSubscriptionNotConfirmed e) {
				logger.warning("Senden der Quittung unter Aspekt '" + qAspId + "' fehlgeschlagen: " + e.getMessage() + Debug.NEWLINE + Util.rd2Str(rd));
			}
		}
	}

	protected void handleErr(Exception e, String corpusDelicti) {
		TOTAL_DS_FAIL.incrementAndGet();
		logger.error(getName() + ": Fehler beim Archivieren eines Datensatzes: " + e.getMessage() + Debug.NEWLINE + Util.rd2Str(rd), e);
	}

	protected void handleErr(Exception e, CloseContainerObject cco) {
		logger.error(getName() + ": Fehler beim Abschliessen eines Containers: " + e.getMessage() + Debug.NEWLINE + cco, e);
	}

	protected void handleSevereErr(Throwable e, String corpusDelicti) {
		TOTAL_DS_FAIL.incrementAndGet();
		logger.error(getName() + ": Schwerer Fehler beim Archivieren eines Datensatzes:" + Debug.NEWLINE + Util.rd2Str(rd), e);	// inkl. Stack Trace
	}

	/**
	 * Erzeugt die Kennzeichen-Datei, ob die Indexe veraendert wurden, um nach einem Absturz den Wiederanlauf zu beschleunigen.
	 *
	 * @throws PersistenceException
	 */
	protected void createRebuildIndexFlagFile() throws PersistenceException {
		if(!din.isRebuildIndexFileFlagCreated(adk)) {
			persMgr.createRebuildIxdFlagFile(didPath);
			din.rebuildIndexFileFlagWasCreated(adk);
		}
	}

	protected void logReceivedData(ResultData rd) {
		logger.finer(getName() + ": Datensatz zur Archivierung erhalten: ", rd);
	}

	protected void logWriteData() throws IndexException {
		logger.finest(
				getName() + ": Datensatz (" + Util.dsKeys2Str(dataIdx, arsTime, dataTime) + ") in Container (ID=" + din.getOpenContID(adk) + ") angefuegt"
		);
	}

	protected void logNewContainer() {
		logger.finer(getName() + ": neuen Container (ID=" + openContID + ") fuer " + Util.did2Str(objID, atg.getId(), aspID, sv, adk) + "angelegt");
	}

	private void logDataGap() {
		logger.fine(
				"Potentielle Datenluecke vermerkt: letzer archivierter Index (O" + (lastADataIdx > lastNDataIdx ? "A" : "N") + "): " + lastDataIdx + " ["
				+ Util.dIdx2Str(lastDataIdx) + "] [" + Util.dIdx2StrExt(lastDataIdx) + "]" + ", erhaltener Datenindex" + dataIdx + " [" + Util.dIdx2Str(dataIdx)
				+ "] [" + Util.dIdx2StrExt(dataIdx) + "]"
		);
	}

	protected void logManualCloseContainer(long closeObjID, CloseContainerObject cco) {
		logger.finer(
				getName() + ": Container (ID=" + openContID + ") fuer " + Util.did2Str(closeObjID, cco.atgID, cco.aspID, sv, adk)
				+ " wg. Parameteraenderung abgeschlossen"
		);
	}

	/**
	 * Prueft ob ein Container abgeschlossen werden muss. Die Pruefung basiert auf {@link #containerSettings} und dem mit {@link
	 * de.bsvrz.ars.ars.persistence.PersistenceManager#accessContainer()} im Zugriff befindlichen Daten. Der Aufrufer dieser Methode muss die Synchronisierung
	 * ueber den DataIdentNode durchfuehren.
	 *
	 * @param requested Wahr fuer nachgeforderte, falsch fuer aktuelle Daten.
	 *
	 * @return Wahr, falls Container abgeschlossen werden muss, falsch sonst.
	 */
	protected boolean mustCloseContainer(boolean requested) throws PersistenceException, IndexException {
		if(din.maxDIdx(adk) > dataIdx) {
			logger.fine(
					getName() + ": Datenindex-Ruecksprung in " + (requested ? "nachgeforderten" : "aktuellen") + " Daten fuer "
					+ Util.did2Str(objID, atg.getId(), aspID, sv, adk) + ": " + Util.timestrMillisFormatted(dataTime) + " < "
					+ Util.timestrMillisFormatted(din.maxDTime(adk))
			);
			return true;
		}
		if(din.maxDTime(adk) > dataTime) {
			logger.fine(
					getName() + ": Datenzeit-Ruecksprung in " + (requested ? "nachgeforderten" : "aktuellen") + " Daten fuer "
					+ Util.did2Str(objID, atg.getId(), aspID, sv, adk) + ": " + Util.timestrMillisFormatted(dataTime) + " < "
					+ Util.timestrMillisFormatted(din.maxDTime(adk))
			);
			return true;
		}
		CloseCondition closeCond = containerSettings.getExceptionSettings(atg);
		if(closeCond != null) {
			return checkCloseCondition(closeCond);
		}
		else {
			return checkCloseCondition(containerSettings.stdCloseConditions);
		}
	}

	/**
	 * Prueft extra auf Containergroesse. Damit kann verhindert werden, dass staendig neue Container angelegt werden, wenn bereits mit einem Datum die Groesse
	 * ueberschritten wuerde. Stattdessen wird mit der maximal erlaubten Groesse eine Fehlermeldung ausgegeben. Der Aufrufer dieser Methode muss die
	 * Synchronisierung ueber den DataIdentNode durchfuehren.
	 *
	 * @return wahr, falls die maximale Containergroesse ueberschritten ist, falsch sonst
	 *
	 * @throws PersistenceException
	 */
	protected boolean tooLargeForNewContainer() throws PersistenceException {
		long totalLen = ContainerFile.getHeaderLen() + dataEntryLen;
		CloseCondition closeCond = containerSettings.getExceptionSettings(atg);
		if(closeCond != null && totalLen > closeCond.maxContSize) {
			logger.warning(
					getName() + ": Datensatz (" + dataEntryLen + " Bytes) + Header (" + ContainerFile.getHeaderLen()
					+ " Bytes) ist grer als die maximale Containerlnge (" + closeCond.maxContSize + " Bytes) +  und wurde nicht archiviert"
			);
			return true;
		}
		else if(totalLen > containerSettings.stdCloseConditions.maxContSize) {
			logger.warning(
					getName() + ": Datensatz (" + dataEntryLen + " Bytes) + Header (" + ContainerFile.getHeaderLen()
					+ " Bytes) ist grer als die maximale Containerlnge (" + containerSettings.stdCloseConditions.maxContSize
					+ " Bytes) +  und wurde nicht archiviert"
			);
			return true;
		}
		return false;
	}

	/**
	 * Prueft ob der aktuelle Container die angegebenen Kriterien zum Container-Abschliessen erfuellt.
	 *
	 * @param closeCond Container-Abschluss-kriterien
	 *
	 * @return Wahr, wenn der Container abgeschlossen werden muss, falsch sonst.
	 *
	 * @throws PersistenceException
	 * @throws IndexException
	 */
	private boolean checkCloseCondition(CloseCondition closeCond) throws PersistenceException, IndexException {
		return (din.getCurContEntries(adk) >= closeCond.maxContAnzDS) || (dataTime - din.minDTime(adk) >= closeCond.maxContTime * 1000) || (
				persMgr.getContainerSize(this) + dataEntryLen > closeCond.maxContSize); // erfordert die meiste Zeit
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Class: CloseContainerObject
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/** Eine Instanz dieser Klasse wird eingefuegt, wenn ein Container aufgrund einer Parametrierungs-Aenderung abgeschlossen werden soll. */
	public static class CloseContainerObject {

		public long atgID, aspID;

		public short sv;

		public CloseContainerObject(DataDescription dd) {
			atgID = dd.getAttributeGroup().getId();
			aspID = dd.getAspect().getId();
			sv = dd.getSimulationVariant();
		}

		public String toString() {
			return "atgID=" + atgID + ", aspID=" + aspID + ", sv=" + sv;
		}
	}
}
