/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.LongTermTaskStatePublisher;
import de.bsvrz.ars.ars.mgmt.datatree.DIdNodeNotFoundException;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.tasks.tasklistener.RingOfDeathListener;
import de.bsvrz.ars.ars.persistence.ContainerFile;
import de.bsvrz.ars.ars.persistence.ContainerFileDirIterator;
import de.bsvrz.ars.ars.persistence.ContainerHdr;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.index.IndexResult;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dataSerializer.Serializer;
import de.bsvrz.sys.funclib.dataSerializer.SerializingFactory;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.MessageType;
import de.bsvrz.sys.funclib.losb.datk.PidQuery;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.util.*;

/**
 * Task zur Bearbeitung von Loeschzeit-Verlaengerungsauftraegen.
 *
 * @author beck et al. projects GmbH
 * @author Christian Wied
 * @version $Revision: 9877 $ / $Date: 2012-01-26 15:00:50 +0100 (Thu, 26 Jan 2012) $ / ($Author: rs $)
 */
public class LZVTask extends SingleClientSenderTask {

	// nur fuer testzwecke
	private static int CAPA_CHECK_RUNS = 0;

	public static final long CHANGETIME_SPONTAN = -1;

	public static final long CHANGETIME_REGULAR = 0;

	private static long MIN_REM_DSK_CAPA = 2 * 1000 * 1000 * 1000; // Default 2GB gemaess DatK

	private static CapacityCheckResultData capacityCheckResultData;

	private static DeleteRegularResultData deleteRegularResultData;

	private byte[] indexMem;

	private int numOfVisitedCont, numOfDeletedCont;

	private LongTermTaskStatePublisher.Task _longTermTaskStatePublisherTask;

	/**
	 * Fr spontantes Lschen: 1. Vorhaltezeitraum abgelaufen, und schon gesichert 2. wiederhergestellte Daten 3. Simulationsdaten 4. Vorhaltezeitraum noch nicht
	 * abgelaufen, aber schon gesichert 5. Lschen durch schrittweises Verschieben des Vorhaltezeitraums
	 * <p/>
	 * Die Reihenfolge der Deklaration ist von Bedeutung!
	 */
	private enum DELETE_LEVEL {

		SAVED_EXPIRED("1 (Vorhaltezeitraum abgelaufen, und schon gesichert)"),
		RESTORED("2 (wiederhergestellte Daten)"),
		SIMULATION("3 (Simulationsdaten)"),
		SAVED_NOTEXPIRED("4 (Vorhaltezeitraum noch nicht abgelaufen, aber schon gesichert)"),
		TIME_SHIFTING("5 (Lschen durch schrittweises Verschieben des Vorhaltezeitraums)");

		private final String _name;

		DELETE_LEVEL(final String name) {
			_name = name;
		}


		public String toString() {
			return _name;
		}
	}

	;

	/** Um wieviele Monate maximal der Vorhaltezeitraum angepasst wird. */
	private static int MAX_TIMESHIFTING = 2 * 12; //

	/**
	 * Erzeugt eine neue Loeschzeit-Verlaengerungs-Task.
	 *
	 * @param archiveMgr Der Archiv-Manager.
	 * @param rodl       Der Listener wird benachrichtigt beim Beenden der Task.
	 * @param inputData  Eingangswarteschlange.
	 *
	 * @throws ConfigurationException
	 */
	public LZVTask(ArchiveManager archiveMgr, RingOfDeathListener rodl, RingBuffer<ResultData> inputData, final LongTermTaskStatePublisher longTermTaskStatePublisher) throws ConfigurationException {
		super(archiveMgr, rodl, inputData);
		setName("LZV-Task");
		logger = Debug.getLogger();
		_longTermTaskStatePublisherTask = longTermTaskStatePublisher.registerTask("Lschen");
	}

	@Override
	public void run() {
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		try {
			super.run();
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.TERMINATED);
		}
	}

	/**
	 * Liefert die Mindest-Restkapazitaet der Festplatte.
	 *
	 * @return Mindestrestkapazitaet in Byte
	 */
	public static long getMinRemainingDiskCapa() {
		return MIN_REM_DSK_CAPA;
	}

	/**
	 * Setzt die Mindest-Restkapazitaet der Festplatte.
	 *
	 * @param minRemainingDiskCapa Neue Mindestrestkapazitaet in Byte.
	 */
	public static void setMinRemainingDiskCapa(long minRemainingDiskCapa) {
		MIN_REM_DSK_CAPA = minRemainingDiskCapa;
	}


	/**
	 * Fhrt das Lschen, spontanes Lschen oder die Lschzeitverlngerung aus.
	 *
	 * @see de.bsvrz.ars.ars.mgmt.tasks.SingleTask#work(ResultData)
	 */
	protected void work(ResultData resultData) {
		if(!resultData.hasData()) return;
		try {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.START_WORK);
			indexMem = new byte[INDEX_MEMORY_SIZE];
			init(resultData.getData());
			changeDeletionTime(resultData);
			if(!(resultData instanceof CapacityCheckResultData) && !(resultData instanceof DeleteRegularResultData)) sendSuccessResponse();
			logger.info(getName() + ": " + numOfVisitedCont + " Container besucht, " + numOfDeletedCont + " davon betroffen");
		}
		catch(Exception e) {
			logger.error("Loeschzeitverlaengerung fehlgeschlagen", e);
			sendErrorResponse("Loeschzeitverlaengerung fehlgeschlagen: " + e.getMessage());
		}
		finally {
			indexMem = null;
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		}
	}


	/**
	 * Prft, welche Operation (Lschen, spontanes Lschen oder die Lschzeitverlngerung) ausgefhrt werden soll, und fhrt diese dann durch.
	 *
	 * @param resultData Datensatz mit Informationen ber den Auftrag
	 *
	 * @throws Exception Fehler
	 */
	void changeDeletionTime(ResultData resultData) throws Exception {
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.WORKING);
		try{
			long changeTime = deserializer.readLong();
			ArchiveDataSpecification[] ads = new ArchiveDataSpecification[0];
			if(!(resultData instanceof CapacityCheckResultData) && !(resultData instanceof DeleteRegularResultData)) {
				// nur externe Auftrge werden geparst
				ads = parseArchiveDataSpec(deserializer, resultData);
			}

			if(changeTime == CHANGETIME_SPONTAN) {
				deleteSpontan(ads, changeTime);
			}

			else if(changeTime == CHANGETIME_REGULAR) {
				long t0 = System.currentTimeMillis();
				logger.info("Regulres Lschen");
				containerWalk(ads, changeTime, null, null);
				logger.info("Regulres Lschen beendet, Dauer: " + ((System.currentTimeMillis() - t0) / 1000) + " Sekunden");
			}

			else if(changeTime > 0) {
				containerWalk(ads, changeTime, null, null);
			}

			else {
				logger.warning(getName() + ": Unbekannte Loeschzeit: " + changeTime);
				throw new Exception(getName() + ": Loeschzeitverlaengerung: Unbekannte Loeschzeit: " + changeTime);
			}
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.FINISH_WORK);
		}
	}

	/**
	 * Der <code>DeleteSpontanTask</code> lscht bei knapp werdendem Speicherplatz unwichtige Container.
	 * <p/>
	 * Dazu ueberprueft er in einem parametrierbaren Zyklus, wieviel Speicherplatz dem Archivsystem noch zur Verfuegung steht. Erreicht die vorhandene Kapazitt
	 * einen Schwellwert, werden folgende Daten geloescht: <ul> <li>archivierte Daten, deren Loeschschutz abgelaufen ist und die nicht (mehr) gesichert werden
	 * muessen,</li> <li>wiederhergestellte Daten,</li> <li>archivierte Simulationsdaten,</li> <li>archivierte Daten, deren Loeschschutz nicht abgelaufen ist, die
	 * aber schon gesichert wurden.</li> </ul>
	 */
	protected void deleteSpontan(ArchiveDataSpecification[] ads, long changeTime) throws Exception {
		long freeDiskSpace = Util.calcFreeDiskSpace(persMgr.getArchivePath());
		Calendar compareDate = Calendar.getInstance();
		boolean someDataDeleted = false;

		for(DELETE_LEVEL level : DELETE_LEVEL.values()) {
			if(freeDiskSpace < getMinRemainingDiskCapa()) {
				someDataDeleted = true;
				switch(level) {
					// Verschieben des Vorhaltezeitraums
					case TIME_SHIFTING:
						for(int i = 0; i < MAX_TIMESHIFTING; i++) {
							compareDate.add(Calendar.MONTH, 1);
							if(freeDiskSpace < getMinRemainingDiskCapa()) {
								long t0 = System.currentTimeMillis();
								logger.info(
										"Spontanes Lschen, Level " + level + " mit " + (i + 1) + " Monaten Verkrzung des Vorhaltezeitraums: "
										+ "Verbleibender Plattenplatz in Bytes: " + freeDiskSpace
								);
								containerWalk(ads, changeTime, compareDate, level);
								freeDiskSpace = Util.calcFreeDiskSpace(persMgr.getArchivePath());
								logger.info(
										"Nach spontanem Lschen, Level " + level + " mit " + (i + 1) + " Monaten Verkrzung des Vorhaltezeitraums: "
										+ "Verbleibender Plattenplatz in Bytes: " + freeDiskSpace + ", " + "Dauer: "
										+ ((System.currentTimeMillis() - t0) / 1000) + " Sekunden"
								);
							}
							else {
								MessageSender.getInstance().sendMessage(
										de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
										MessageGrade.INFORMATION,
										"Das spontane Lschen hat Daten gelscht"
								);
								logger.info("Das spontane Lschen hat Daten gelscht");
								return;
							}
						}
						break;

						// Lschung nach den 4 Kriterien in der TAnf 5.1.2.7
					default:
						long t0 = System.currentTimeMillis();
						logger.info("Spontanes Lschen, Level " + level + ": Verbleibender Plattenplatz in Bytes: " + freeDiskSpace);
						containerWalk(ads, changeTime, compareDate, level);
						freeDiskSpace = Util.calcFreeDiskSpace(persMgr.getArchivePath());
						logger.info(
								"Nach spontanem Lschen, Level " + level + ": Verbleibender Plattenplatz in Bytes: " + freeDiskSpace + ", " + "Dauer: "
								+ ((System.currentTimeMillis() - t0) / 1000) + " Sekunden"
						);
				}
			}
			else {
				if(someDataDeleted) {
					MessageSender.getInstance().sendMessage(
							de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
							MessageGrade.INFORMATION,
							"Das spontane Lschen hat Daten gelscht"
					);
					logger.info("Das spontane Lschen hat Daten gelscht");
				}
				return; // Es wurde genug PLatz freigegeben -> fertig
			}
		}

		// Wenn jetzt noch immer nicht genug Platz frei ist, belegt u.U. die Verwaltungsinfo
		// alleine schon so viel, dass wir durch Lschen keinen Platz mehr bekommen
		MessageSender.getInstance().sendMessage(
				de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
				MessageGrade.WARNING,
				"Achtung: Der Speicherplatz geht zu Ende! Durch spontanes Lschen konnte nicht so viel Platz freigegeben werden, wie laut Parametrierung mindestens frei sein soll."
		);
		logger.warning(
				"Achtung: Der Speicherplatz geht zu Ende! Durch spontanes Lschen konnte nicht so viel Platz freigegeben werden, wie laut Parametrierung mindestens frei sein soll."
		);
	}


	/**
	 * Die generelle Container-Walk-Methode.
	 * <p/>
	 * Je nachdem, ob die ArchiveDataSpec gefuellt ist oder nicht, wird ein ContainerWalk ber das komplette Persistenz-Verzeichnis (containerWalkAllFromDir) oder
	 * alle spezifizierten Container (containerWalkAllByDataSpec) ausgefhrt.<br><br>
	 * <p/>
	 * Achtung!<br>Die Methode lscht je nach bergebenen Parametern Archivdaten!
	 *
	 * @param ads
	 * @param changeTime
	 * @param compareDate
	 *
	 * @throws PersistenceException
	 * @throws IndexException
	 */
	void containerWalk(ArchiveDataSpecification[] ads, long changeTime, Calendar compareDate, DELETE_LEVEL level) throws Exception {
		numOfVisitedCont = numOfDeletedCont = 0;
		if(ads.length == 0) {
			containerWalkAllFromDir(changeTime, compareDate, level);
		}
		else {
			containerWalkByArchiveDataSpec(ads, changeTime, compareDate, level);
		}
	}

	private void containerWalkAllFromDir(long changeTime, Calendar compareDate, DELETE_LEVEL level) throws Exception {
		File didDir;
		ContainerFileDirIterator iter = new ContainerFileDirIterator(persMgr.getArchivePath());
		while((didDir = iter.getNextDIdDirectory()) != null) {
			long[] did = PersistenceManager.extractDIdFromPath(didDir);
			ArchiveDataKind adk = PersistenceManager.getArchiveDataKindFromDir(didDir);

			File[] files = didDir.listFiles(ContainerFile.CONT_FILENAME_FILTER);
			if(files != null) {
				for(File cont : files) {
					if(isTaskTerminated()) throw new Exception(getName() + ": Archivsystem wurde beendet.");

					DataIdentNode din = didTree.existsCreate(did[0], did[1], did[2], (short)did[3]); // nie null
					boolean busyAreaEntered = false;
					try {
						//Prfen, ob Archivsystem berlastet und Task deshalb angehalten werden soll
						//Spontanes Lschen darf aber nicht angehalten werden, sondern nur Lschzeitverlngerung und (nicht-spontanes) Lschen.
						if(changeTime != CHANGETIME_SPONTAN) suspendTaskIfNecessary();

						// Synchronisierung ueber DataIdentNode:
						busyAreaEntered = din.enterBusyArea();
						if(!busyAreaEntered) {
							logger.warning(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
							throw new Exception(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
						}
						// Auf Container zugreifen und bearbeiten:
						persMgr.accessContainer(this, did[0], did[1], did[2], (short)did[3], adk, PersistenceManager.getContID(cont));
						numOfVisitedCont++;
						handleContainer(changeTime, compareDate, level);
					}
					catch(PersistenceException e) {
						throw e;
					}
					finally {
						persMgr.leaveContainer(this);
						if(din != null && busyAreaEntered) din.exitBusyArea();
					}
				}
			}
		}
	}

	private void containerWalkByArchiveDataSpec(ArchiveDataSpecification[] ads, long changeTime, Calendar compareDate, DELETE_LEVEL level) throws Exception {

		for(int i = 0; i < ads.length; i++) {
			ArchiveDataSpecification arcDataSpec = ads[i];
			DataIdentNode din = didTree.existsCreate(arcDataSpec); // DataIdentNode ist hier nie null

			boolean criticalAreaEntered = false;
			try {
				// Synchronisierung ueber DataIdentNode:
				criticalAreaEntered = din.enterCriticalArea();
				if(!criticalAreaEntered) {
					logger.warning(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
					throw new Exception(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
				}

				// Indexe aller Datensatzarten anlegen:
				// (Alle, damit die Container all der Datensatzarten behandelt werden koennen, fuer die Daten vorhanden sind.)
				for(ArchiveDataKind dataKind : Util.getDataKinds(arcDataSpec.getDataKinds())) {
					if(!din.indexesAssigned(dataKind)) {
						try {
							persMgr.assignIndexes(didTree, arcDataSpec, dataKind, this, indexMem);
						}
						catch(DIdNodeNotFoundException e) {
						} // Exception tritt nicht auf, da DIN oben angelegt worden ist.
					}
				}
			}
			finally {
				if(din != null && criticalAreaEntered) din.exitCriticalArea();
			}

			ArchiveTimeSpecification atSpec = arcDataSpec.getTimeSpec();
			long start = atSpec.getIntervalStart();
			long end = atSpec.getIntervalEnd();

			for(ArchiveDataKind dataKind : Util.getDataKinds(arcDataSpec.getDataKinds())) {
				if(isTaskTerminated()) throw new Exception(getName() + ": Archivsystem wurde beendet.");

				IndexResult ir;
				criticalAreaEntered = false;
				try {
					// Synchronisierung ueber DataIdentNode:
					criticalAreaEntered = din.enterCriticalArea();
					if(!criticalAreaEntered) {
						logger.warning(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
						throw new Exception(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
					}
					// Pruefen, ob jemals Daten fuer diese Datenidentifikation und Datensatzart archiviert wurden:
					if(atSpec.getTimingType().equals(TimingType.DATA_INDEX)) {
						ir = din.getDataIdxContIDs(dataKind, persMgr.getPathName(arcDataSpec, dataKind), indexMem, start, end);
					}
					else if(atSpec.getTimingType().equals(TimingType.DATA_TIME)) {
						ir = din.getDataTimeContIDs(dataKind, persMgr.getPathName(arcDataSpec, dataKind), indexMem, start, end);
					}
					else if(atSpec.getTimingType().equals(TimingType.ARCHIVE_TIME)) {
						ir = din.getArSTimeContIDs(dataKind, persMgr.getPathName(arcDataSpec, dataKind), indexMem, start, end);
					}
					else {
						return;
					}
				}
				finally {
					if(din != null && criticalAreaEntered) din.exitCriticalArea();
				}

				// Alle ueber die Indexe gefundenen Container behandeln:
				for(int j = 0; j < ir.size(); j++) {
					boolean busyAreaEntered = false;
					try {
						//Prfen, ob Archivsystem berlastet und Task deshalb angehalten werden soll
						//Spontanes Lschen darf aber nicht angehalten werden, sondern nur Lschzeitverlngerung und (nicht-spontanes) Lschen.
						if(changeTime != CHANGETIME_SPONTAN) suspendTaskIfNecessary();

						// Synchronisierung ueber DataIdentNode:
						busyAreaEntered = din.enterBusyArea();
						if(!busyAreaEntered) {
							logger.warning(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
							throw new Exception(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
						}
						// Auf Container zugreifen und bearbeiten:
						persMgr.accessContainer(this, persMgr.getPathName(arcDataSpec, dataKind), ir.getCID(j));
						numOfVisitedCont++;
						handleContainer(changeTime, compareDate, level);
					}
					finally {
						persMgr.leaveContainer(this);
						if(din != null && busyAreaEntered) din.exitBusyArea();
					}
				}
			}
		}
	}

	/**
	 * Bearbeitet den individuellen Container.
	 * <p/>
	 * Fhrt Fallunterscheidung auf der <code>changeTime</code> aus und ruft die 3 Loesch-Methoden auf.
	 *
	 * @param changeTime
	 * @param compareDate
	 *
	 * @throws PersistenceException
	 */
	private void handleContainer(long changeTime, Calendar compareDate, DELETE_LEVEL level) throws PersistenceException {
		if(changeTime == CHANGETIME_SPONTAN) {
			checkAndDeleteContainerSpontan(this, persMgr, compareDate, level);
			_longTermTaskStatePublisherTask.setTaskProgress("Spontanes Lschen: %1$d Container untersucht, %3$d Container gelscht", numOfVisitedCont, -1, numOfDeletedCont);
		}
		else if(changeTime == CHANGETIME_REGULAR) {
			checkAndDeleteContainerRegular(this, persMgr);
			_longTermTaskStatePublisherTask.setTaskProgress("Regulres Lschen: %1$d Container untersucht, %3$d Container gelscht", numOfVisitedCont, -1, numOfDeletedCont);
		}
		else {
			persMgr.readContainerHeader(this);
			long loeschschutz = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_LOESCHUTZ);
			persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_LOESCHUTZ, loeschschutz + changeTime);
			persMgr.writeContainerHeader(this);
			numOfDeletedCont++;
			_longTermTaskStatePublisherTask.setTaskProgress("Lschschutzverlngerung: %1$d Container untersucht, %3$d Container angepasst", numOfVisitedCont, -1, numOfDeletedCont);
		}
	}

	/**
	 * Prueft, ob der Container existiert.<br> Wenn ja, wird geprueft, ob der Container zu loeschen ist. <br> Wenn ja, wird der Container geloescht.
	 *
	 * @param compareDate
	 *
	 * @throws PersistenceException
	 */
	void checkAndDeleteContainerSpontan(Task task, PersistenceManager persMgr, Calendar compareDate, DELETE_LEVEL level) throws PersistenceException {
		if(persMgr.existsContainer(task) && hasToBeDeletedSpontan(task, persMgr, compareDate, level)) {
			persMgr.deleteContainerData(task);
			persMgr.setContainerHeaderParam(task, ContainerHdr.CHP_DELETED, "1");
			persMgr.writeContainerHeader(this);
			numOfDeletedCont++;
			logger.fine("DeleteSpontan: Container mit ID <" + persMgr.getContainerHeaderParamAsString(task, ContainerHdr.CHP_CONT_ID) + "> gelscht.");
		}
	}

	/**
	 * Prueft, ob das ContainerFile geloescht werden muss. Folgende Faktoren werden betrachtet: <ul> <li>archivierte Daten, deren Loeschschutz abgelaufen ist und
	 * die nicht (mehr) gesichert werden muessen,</li> <li>wiederhergestellte Daten,</li> <li>archivierte Simulationsdaten,</li> <li>archivierte Daten, deren
	 * Loeschschutz nicht abgelaufen ist, die aber schon gesichert wurden.</li> </ul>
	 *
	 * @param compareDate
	 *
	 * @return boolean
	 *
	 * @throws PersistenceException
	 */
	boolean hasToBeDeletedSpontan(Task task, PersistenceManager persMgr, Calendar compareDate, DELETE_LEVEL level) throws PersistenceException {
		try {
			persMgr.readContainerHeader(task);
			// aktueller, offener Container wird nicht gelscht
			if(persMgr.getContainerHeaderParamAsInt(task, ContainerHdr.CHP_ANZ_DS) < 0) return false;

			// schon gelschte Container nicht nochmal lschen
			if(persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_DELETED)) return false;

			// Kein level gesetzt -> keine weiteren Checks zum Lschen
			if(level == null) return false;

			switch(level) {
				case SAVED_EXPIRED:
				case TIME_SHIFTING:
					// zu sichernder, aber noch nicht gesicherter, Container wird nicht gelscht
					if(persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_TO_SAVE) && persMgr.getContainerHeaderParamAsString(
							task, ContainerHdr.CHP_MEDIUM_ID
					).equals(ContainerHdr.MEDIUM_UNSAVED)) {
						return false;
					}
					// TAnf 5.1.2.7 erster Absatz: (gesichert || nicht_zu_sichern) && Vorhaltezeitraum_abgelaufen && (kein_Lschschutz || Lschschutz_abgelaufen)
					long chpLoeschutz = persMgr.getContainerHeaderParamAsLong(task, ContainerHdr.CHP_LOESCHUTZ);

					long delAt = persMgr.getContainerHeaderParamAsLong(task, ContainerHdr.CHP_LOESCHEN);
					long compareDateTimeInMillis = compareDate.getTimeInMillis();
					return delAt < compareDateTimeInMillis && (chpLoeschutz == 0L || chpLoeschutz < compareDateTimeInMillis);
				case RESTORED:
					if(persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_RESTORED)) return true;
				case SIMULATION:
					if(persMgr.getContainerHeaderParamAsInt(task, ContainerHdr.CHP_SIM_VAR) != 0) return true;
				case SAVED_NOTEXPIRED:
					if(!persMgr.getContainerHeaderParamAsString(task, ContainerHdr.CHP_MEDIUM_ID).equals("unsaved")) return true;
			}
		}
		catch(Exception e) {
			logger.warning("Problem bei der berprfung, ob ein Container spontan gelscht werden kann", e);
		}

		return false;
	}

	/** Prueft, ob der Container existiert.<br> Wenn ja, wird geprueft, ob der Container zu loeschen ist. <br> Wenn ja, wird der Container geloescht. */
	void checkAndDeleteContainerRegular(Task task, PersistenceManager persMgr) throws PersistenceException {
		if(persMgr.existsContainer(task) && hasToBeDeletedRegular(task, persMgr)) {
			persMgr.deleteContainerData(task);
			persMgr.setContainerHeaderParam(task, ContainerHdr.CHP_DELETED, "1");
			persMgr.writeContainerHeader(this);
			numOfDeletedCont++;
			logger.fine("DeleteRegular: Container with id <" + persMgr.getContainerHeaderParamAsString(task, ContainerHdr.CHP_CONT_ID) + "> deleted.");
		}
	}

	/**
	 * Prueft, ob das ContainerFile geloescht werden muss.
	 *
	 * @return boolean
	 *
	 * @throws PersistenceException
	 */
	boolean hasToBeDeletedRegular(Task task, PersistenceManager persMgr) throws PersistenceException {
		try {
			persMgr.readContainerHeader(task);
			// aktueller, offener Container wird nicht gelscht
			if(!persMgr.isContainerClosed(task)) return false;

			// zu sichernder, aber noch nicht gesicherter, Container wird nicht gelscht
			if(persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_TO_SAVE) && persMgr.getContainerHeaderParamAsString(
					task, ContainerHdr.CHP_MEDIUM_ID
			).equals(ContainerHdr.MEDIUM_UNSAVED)) {
				return false;
			}

			// schon gelschte Container nicht nochmal lschen
			if(persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_DELETED)) return false;

			// TAnf 5.1.2.7 erster Absatz: (gesichert || nicht_zu_sichern) && Vorhaltezeitraum_abgelaufen && (kein_Lschschutz || Lschschutz_abgelaufen)
			long chpLoeschutz = persMgr.getContainerHeaderParamAsLong(task, ContainerHdr.CHP_LOESCHUTZ);
			long now = System.currentTimeMillis();

			if(!persMgr.getContainerHeaderParamAsString(task, ContainerHdr.CHP_MEDIUM_ID).equals(ContainerHdr.MEDIUM_UNSAVED)
			   || !persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_TO_SAVE)) {

				long delAt = persMgr.getContainerHeaderParamAsLong(task, ContainerHdr.CHP_LOESCHEN);
				if(delAt < now && (chpLoeschutz == 0L || chpLoeschutz < now)) return true;
			}
		}
		catch(Exception e) {
			logger.warning("Problem bei der berprfung, ob ein Container spontan gelscht werden kann", e);
		}

		// Sonst nichts lschen
		return false;
	}

	public static CapacityCheckResultData getCapacityCheckResultData(ArchiveManager archMgr) {
		if(capacityCheckResultData == null) {
			Data data = archMgr.getDavCon().createData(archMgr.getAtg("atg.archivAnfrageSchnittstelle"));
			data.getReferenceValue(PidQuery.ATT_SENDER_NAME).setSystemObject(archMgr.getDavCon().getLocalApplicationObject());
			data.getItem(PidQuery.ATT_QUERY_IDX_NAME).asUnscaledValue().set(0);
			data.getUnscaledValue(PidQuery.ATT_MESSAGE_TYP_NAME).set(MessageType.DELETION_TIME);
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			try {
				Serializer serializer = SerializingFactory.createSerializer(bos);
				serializer.writeInt(serializer.getVersion());
				serializer.writeLong(CHANGETIME_SPONTAN);
			}
			catch(Exception e) {
//                logger.warning("Konnte lzvResultData nicht richtig initialisieren: ", e.getMessage());  // TODO Auskommentierter Code !?
			}
			byte b[] = bos.toByteArray();
			SerializerUtil.insertArray(data.getArray(PidQuery.ATT_DATA_NAME), b, b.length);

			capacityCheckResultData = new CapacityCheckResultData(null, null, 0, data);
		}
		return capacityCheckResultData;
	}

	public static DeleteRegularResultData getDeleteRegularResultData(ArchiveManager archMgr) {
		if(deleteRegularResultData == null) {
			Data data = archMgr.getDavCon().createData(archMgr.getAtg("atg.archivAnfrageSchnittstelle"));
			data.getReferenceValue(PidQuery.ATT_SENDER_NAME).setSystemObject(archMgr.getDavCon().getLocalApplicationObject());
			data.getItem(PidQuery.ATT_QUERY_IDX_NAME).asUnscaledValue().set(0);
			data.getUnscaledValue(PidQuery.ATT_MESSAGE_TYP_NAME).set(MessageType.DELETION_TIME);
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			try {
				Serializer serializer = SerializingFactory.createSerializer(bos);
				serializer.writeInt(serializer.getVersion());
				serializer.writeLong(CHANGETIME_REGULAR);
			}
			catch(Exception e) {
//                logger.warning("Konnte lzvResultData nicht richtig initialisieren: ", e.getMessage()); // TODO Auskommentierter Code !?
			}
			byte b[] = bos.toByteArray();
			SerializerUtil.insertArray(data.getArray(PidQuery.ATT_DATA_NAME), b, b.length);

			deleteRegularResultData = new DeleteRegularResultData(null, null, 0, data);
		}
		return deleteRegularResultData;
	}

	public static class CapacityCheckResultData extends ResultData {

		public CapacityCheckResultData(SystemObject object, DataDescription dataDescription, long time, Data data) {
			super(object, dataDescription, time, data);
		}
	}

	public static class DeleteRegularResultData extends ResultData {

		public DeleteRegularResultData(SystemObject object, DataDescription dataDescription, long time, Data data) {
			super(object, dataDescription, time, data);
		}
	}
}
