/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.util.zip.Inflater;

/**
 * Iterator fuer Datensaetze, die ein Datencontainer auf Anfrage geliefert hat. Mit {@link #readNextDataHdr()} kann geprueft werden, ob ein weiterer Datensatz
 * verfuegbar ist. Falls ja, koennen seine Header-Daten mit {@link #getArcTime()}, etc. eingesehen werden und der Datensatz selbst mit {@link
 * #deserializeData()} abgeholt oder mit {@link #skipData()} uebersprungen werden. <br><br> Beispiel:
 * <blockquote><pre>
 * ContainerDataIterator it = myPersistenceManager.getArchiveData(...);
 * while(it.readNextDataHdr()) {
 *     if (it.getDataIndex() > myBorder) {
 *         Data data = it.deserializeData();
 *     } else {
 *         it.skipData();
 *     }
 * }
 * </pre></blockquote>
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @version $Revision: 6419 $ / $Date: 2009-03-10 23:15:57 +0100 (Tue, 10 Mar 2009) $ / ($Author: rs $)
 */
public class ContainerDataIterator {

	private ContainerFile containerFile;

	private long dataTime, arcTime, dataIndex;

	private boolean isNoSource, isNodata, isPotDataGap, isNoRights, hdrMinVal, hdrMaxVal;

	private int counter, dataSize, dataUncSize;

	private ArchiveDataKind dataKind;

	private boolean useExtraBuffer;

	/**
	 * Erzeugt einen neuen Iterator fuer das angegebene {@link ContainerFile}-Objekt.
	 *
	 * @param containerFile ContainerFile
	 */
	ContainerDataIterator(ContainerFile containerFile) {
		counter = 0;
		this.containerFile = containerFile;
	}

	/**
	 * Prueft, ob ein weiterer Datensatz verfuegbar ist. Wenn dies der Fall ist, wird <code>true</code> zurueck geliefert und die Header-Daten des Datensatzes
	 * koennen mit {@link getArcTime()}, {@link #getDataTime()} und {@link #getDataIndex()} eingesehen werden. Der Datensatz selbst ist ueber die Methode {@link
	 * #deserializeData()} abrufbar, oder kann mit der Methode {@link #skipData()} uebersprungen werden. Wenn kein Datensatz verfuegbar ist, wird
	 * <code>false</code> zurueck geliefert.
	 *
	 * @return Kennzeichen, ob Datensatz verfuegbar ist
	 *
	 * @throws PersistenceException, falls ein Fehler in der Persistenz aufgetreten ist
	 */
	public boolean readNextDataHdr() throws PersistenceException {
		return containerFile.iterReadNextDataHdr();
	}

	public String getHeader() {
		int adkIdx = Util.getDataKindIndex(dataKind);
		return (adkIdx != -1 ? PersistenceManager.DATAKIND_SUFFIX[adkIdx] : "") + " " + ((isData())
		                                                                                 ? ("(" + Util.leadBlank(getDataSize(), 5) + "/"
		                                                                                    + (isCompressed() ? Util.leadBlank(getDataUncSize(), 5) : " --- ")
		                                                                                    + ") ")
		                                                                                 : ((isNoData() ? "[keine Daten] " : "")
		                                                                                    + (isNoSource() ? "[keine Quelle]" : "")
		                                                                                    + (isNoRights() ? "[keine Rechte]" : "") + (isPotDataGap()
		                                                                                                                                ? "[pot. Luecke] "
		                                                                                                                                : ""))) + " dx="
		                                                                                                                                        + Util.dIdx2StrExt(
				getDataIndex()
		) + " (" + getDataIndex() + ")" + " dt=" + Util.timestrMillisFormatted(getDataTime()) + " at=" + Util.timestrMillisFormatted(getArcTime());
	}

	/**
	 * Schliesst den Iterator und verwirft die zu iterierende Ergebnis-Menge.
	 *
	 * @throws PersistenceException, falls ein Fehler in der Persistenz aufgetreten ist
	 */
	public void close() throws PersistenceException {
		containerFile.iterClose();
	}

	/**
	 * Liefert den serialisierten Datensatz dessen Header-Daten zuvor mittels {@link #readNextDataHdr()} eingelesen worden sind. Der Datensatz kann komprimiert
	 * sein.
	 *
	 * @return Datensatz
	 *
	 * @throws PersistenceException, falls ein Fehler in der Persistenz aufgetreten ist
	 */
	public byte[] getRawData(Inflater decompresser) throws PersistenceException {
		return containerFile.iterGetRawData(decompresser);
	}

	/**
	 * Ueberspringt den Datensatz dessen Header-Daten zuvor mittels {@link #readNextDataHdr()} eingelesen worden sind.
	 *
	 * @return Datensatz
	 *
	 * @throws PersistenceException, falls ein Fehler in der Persistenz aufgetreten ist
	 */
	public void skipData() throws PersistenceException {
		containerFile.iterSkipData();
	}

	/** @return Ob der Datensatz im Header als komprimiert gekennzeichnet wurde */
	public boolean isCompressed() {
		return dataUncSize != ContainerFile.NOT_COMPRESSED;
	}

	/**
	 * @return Ob der Datensatz nach dem Lesen mit {@link #getRawData()} immer noch komprimiert ist. Das ArS kann so konfiguriert werden, dass es beim Lesen die
	 *         Datensaetze dekomprimiert und sie unkomprimiert verschickt. Ist deutlich schneller als das Dekomprimieren durch das Stream-Interface.
	 */
	public boolean isStillCompressed() {
		return containerFile.iterStillCompressed();
	}

	/**
	 * Liefert den Archivzeitstempel des Datensatzes dessen Header-Daten zuvor mittels {@link #readNextDataHdr()} eingelesen worden sind.
	 *
	 * @return Archivzeitstempel in Millisekunden seit 1970.
	 */
	public long getArcTime() {
		return arcTime;
	}

	/**
	 * Liefert den Datenindex des Datensatzes dessen Header-Daten zuvor mittels {@link #readNextDataHdr()} eingelesen worden sind.
	 *
	 * @return Datenindex
	 */
	public long getDataIndex() {
		return dataIndex;
	}

	/**
	 * @param tt Datenzeit, Archivzeit, Datenindex
	 *
	 * @return Datenzeit, Archivzeit, Datenindex (je nach Parameter)
	 */
	public long getTimeIndex(TimingType tt) {
		if(tt == TimingType.ARCHIVE_TIME) {
			return arcTime;
		}
		else if(tt == TimingType.DATA_TIME) {
			return dataTime;
		}
		else {
			return dataIndex;
		}
	}

	/** @return Laufende Nummer des Datenindex (ohne Anmeldezeit Quelle und Modifier-Bits) */
	public long getDataIndexLfdNr() {
		return Util.dIdxLfdnr(getDataIndex());
	}

	/**
	 * Liefert den Datenzeitstempel des Datensatzes dessen Header-Daten zuvor mittels {@link #readNextDataHdr()} eingelesen worden sind.
	 *
	 * @return Datenzeitstempel in Millisekunden seit 1970.
	 */
	public long getDataTime() {
		return dataTime;
	}

	/**
	 * Liefert die Markierung fuer diesen Datensatz.
	 *
	 * @return
	 */
	public DataState getDataState() {
		if(isNoData()) {
			return DataState.NO_DATA;
		}
		else if(isNoSource) {
			return DataState.NO_SOURCE;
		}
		else if(isPotDataGap) {
			return DataState.POSSIBLE_GAP;
		}
		else {
			return DataState.DATA;
		}
	}

	/**
	 * Liefert die Position des Datensatzes dessen Header-Daten zuvor mittels {@link #readNextDataHdr()} eingelesen worden sind.
	 *
	 * @return Position (0 = erster)
	 */
	public int getCounter() {
		return counter;
	}


	public boolean isHdrMinVal() {
		return hdrMinVal;
	}

	void setHdrMinVal(boolean hdrMinVal) {
		this.hdrMinVal = hdrMinVal;
	}

	public boolean isHdrMaxVal() {
		return hdrMaxVal;
	}

	void setHdrMaxVal(boolean hdrMaxVal) {
		this.hdrMaxVal = hdrMaxVal;
	}

	/** @return Wahr, falls der Datensatz keine spezielle Markierung besitzt */
	public boolean isData() {
		return !isNodata && !isNoSource && !isPotDataGap;
	}

	/**
	 * Nachdem {@link #readNextDataHdr()} aufgerufen wurde, gibt diese Methode an, ob der eingelesene Datensatz eine "keine Daten"-Markierung besitzt.
	 *
	 * @return Wahr, falls der Datensatz eine "keine Daten"-Markierung besitzt
	 */
	public boolean isNoData() {
		return isNodata;
	}

	/**
	 * Wird von der Persistenz benutzt, um die Markierung "keine Daten" des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param isNodata
	 */
	void setNodata(boolean isNodata) {
		this.isNodata = isNodata;
	}

	/**
	 * Nachdem {@link #readNextDataHdr()} aufgerufen wurde, gibt diese Methode an, ob der eingelesene Datensatz eine "keine Rechte"-Markierung besitzt.
	 *
	 * @return Wahr, falls der Datensatz eine "keine Rechte"-Markierung besitzt
	 */
	public boolean isNoRights() {
		return isNoRights;
	}

	/**
	 * Wird von der Persistenz benutzt, um die Markierung "keine Rechte" des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param isNoRights
	 */
	void setNoRights(boolean isNoRights) {
		this.isNoRights = isNoRights;
	}

	/**
	 * Nachdem {@link #readNextDataHdr()} aufgerufen wurde, gibt diese Methode an, ob der eingelesene Datensatz eine "keine Quelle"-Markierung besitzt.
	 *
	 * @return Wahr, falls der Datensatz eine "keine Quelle"-Markierung besitzt
	 */
	public boolean isNoSource() {
		return isNoSource;
	}

	/**
	 * Wird von der Persistenz benutzt, um die Markierung "keine Quelle" des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param isNoSource
	 */
	void setNoSource(boolean isNoSource) {
		this.isNoSource = isNoSource;
	}

	/**
	 * Nachdem {@link #readNextDataHdr()} aufgerufen wurde, gibt diese Methode an, ob der eingelesene Datensatz als "potentielle Datenluecke" gekennzeichnet
	 * wurde.
	 *
	 * @return Wahr, falls der Datensatz als "potentielle Datenluecke" gekennzeichnet wurde
	 */
	public boolean isPotDataGap() {
		return isPotDataGap;
	}

	/**
	 * Wird von der Persistenz benutzt, um die Markierung "potentielle Datenluecke" des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param isPotDataGap
	 */
	void setPotDataGap(boolean isPotDataGap) {
		this.isPotDataGap = isPotDataGap;
	}

	/**
	 * Wird von der Persistenz benutzt, um den Archivzeitstempel des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param arcTime Archivzeitstempel
	 */
	void setArcTime(long arcTime) {
		this.arcTime = arcTime;
	}

	/**
	 * Wird von der Persistenz benutzt, um den Datenindex des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param dataIndex Datenindex
	 */
	void setDataIndex(long dataIndex) {
		this.dataIndex = dataIndex;
	}

	/**
	 * Wird von der Persistenz benutzt, um den Datenzeitstempel des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param dataTime Datenzeitstempel
	 */
	void setDataTime(long dataTime) {
		this.dataTime = dataTime;
	}


	public ArchiveDataKind getArchiveDataKind() {
		return dataKind;
	}

	void setArchiveDataKind(ArchiveDataKind dataKind) {
		this.dataKind = dataKind;
	}

	public int getDataSize() {
		return dataSize;
	}

	void setDataSize(int dataSize) {
		this.dataSize = dataSize;
	}

	public int getDataUncSize() {
		return dataUncSize;
	}

	void setDataUncSize(int dataUncSize) {
		this.dataUncSize = dataUncSize;
	}

	/**
	 * Wird von der Persistenz benutzt, um die Position des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param counter Position des Datensatzes (0 = erster)
	 */
	void setCounter(int counter) {
		this.counter = counter;
	}

	boolean mustUseExtraBuffer() {
		return useExtraBuffer;
	}

	void setUseExtraBuffer(boolean useExtraBuffer) {
		this.useExtraBuffer = useExtraBuffer;
	}


}
