/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.persistence.index;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.RandomAccessFile;

/**
 * Diese Klasse stellt einen Datenindex-Index dar. Sie verhlt sich hnlich dem {@link DataTimeIndex}. Die Folge der Eintraege (min, max) ist nicht monoton,
 * weil es beim Nachfordern Rueckspruenge geben kann. Stattdessen sind die Eintraege nur nach dem min-Wert sortiert. Ein Eintrag enthaelt Minimum, Maximum des
 * Datenindex und die ContainerID.
 *
 * @author beck et al. projects GmbH
 * @author Phil Schrettenbrunner
 * @version $Revision: 8254 $ / $Date: 2010-10-06 01:15:27 +0200 (Wed, 06 Oct 2010) $ / ($Author: rs $)
 */
public final class DataIndexIndex {

	/** Dateiname, unter dem alle Indices dieser Art gespeichert werden */
	public static final String IDX_FILENAME = "_dataindex.idx";

	/** Flag, mit dem signalisiert wird, dass ein Eintrag "keinen Wert" hat */
	public static final int NO_VALUE = -1;

	/** Modi, mit denen ein Index angelegt werden kann (siehe {@link DataTimeIndex#DataTimeIndex(int,String,byte[],IndexStartmode) } ) */
	public static enum IndexStartmode {

		APPEND_REUSE_LAST,
		ERASE_OLD
	}

	;

	public static final IndexStartmode APPEND_REUSE_LAST = IndexStartmode.APPEND_REUSE_LAST;

	public static final IndexStartmode ERASE_OLD = IndexStartmode.ERASE_OLD;

	/** Positionen einzelner Daten im Enty-Array */
	private static final int MIN_POS = 0;

	private static final int MAX_POS = 1;

	private static final int CONTID_POS = 2;

	private static final String ERROR_OPEN = "Die Indexdatei konnte nicht geffnet werden. ";

	private static final String ERROR_CLOSE = "Die Indexdatei konnte nicht geschlossen werden. ";

	private static final String ERROR_ACCESS = "Fehler beim Zugriff auf die Indexdatei. ";

	/** Hilfsvariablen */
	private long biggestDIMaxInFile = NO_VALUE;

	private long filePositionReactivatedEntry = NO_VALUE;

	/** Current Entry Array - aktuelle Daten */
	private long currentEntry[] = new long[]{NO_VALUE, NO_VALUE, NO_VALUE};

	/**
	 * Erzeugt einen neuen DatenIndex.
	 * <p/>
	 * Bei startmode==APPEND_REUSE_LAST wird der Ringpuffer mit den aktuellsten Daten aus einer vorhandenen Indexdatei gefllt. Der Eintrag mit der grten
	 * ContainerId wird als currentEntry reaktiviert. Damit kann ein Container fortgesetzt und der reaktivierte Eintrag weiter aktualisiert werden.
	 * <p/>
	 * Bei startmode==ERASE_OLD wird ein leerer Index erzeugt und die bisherige Indexdatei (sofern vorhanden) geloescht.
	 *
	 * @param bufferSize Anzahl der Eintrge, die der die der Index im RAM halten soll
	 * @param basePath   Verzeichnis, in dem die Indexdatei liegt
	 * @param temp       Ein temporres ByteArray, das zum Suchen des currentEntries verwendet wird
	 * @param startmode  {@link IndexStartmode#ERASE_OLD}: alte Indexdatei loeschen, {@link IndexStartmode#APPEND_REUSE_LAST}: anfuegen, letzten Eintrag
	 *                   weiterverwenden
	 *
	 * @throws IndexException Bei Problemem mit dem Dateizugriff
	 */
	public DataIndexIndex(int bufferSize, String basePath, byte temp[], IndexStartmode startmode) throws IndexException {
		this(bufferSize);
//        System.out.println(startmode);
		if(startmode == IndexStartmode.ERASE_OLD) {
			File oldIndexFile = new File(basePath, IDX_FILENAME);
			if(oldIndexFile.exists() && !oldIndexFile.delete()) {
				throw new IndexException(
						"Fehler beim Neuanlegen eines Index: Alte Indexdatei '" + oldIndexFile.getAbsolutePath() + "' konnte nicht geloescht werden: "
				);
			}
		}
		else if(startmode == IndexStartmode.APPEND_REUSE_LAST) {
			if(new File(basePath, IDX_FILENAME).exists()) {
				RandomAccessFile raf = ifOpen(basePath);
				try {
					rbLoad(raf);
					long maxima[][] = getMaximaFromFile(raf, temp, 0, (int)raf.length());
					biggestDIMaxInFile = maxima[0][0];
					if(currentEntry[CONTID_POS] < maxima[2][CONTID_POS]) {
						currentEntry = maxima[2];
						filePositionReactivatedEntry = maxima[1][0];
					}
					rbExtractMaxima();
				}
				catch(Exception e) {
					throw new IndexException(ERROR_ACCESS + e.getMessage(), e);
				}
				finally {
					ifClose(raf);
				}
			}
		}
	}


	/**
	 * Erzeugt einen neuen DatenIndexIndex. Eine eventuell vorhandene Datei wird weiterbenutzt, ein weiterbenutzen des currentEntries ist nicht mglich.
	 *
	 * @param bufferSize Anzahl der Eintrge, die der die der Index im RAM halten soll
	 */
	public DataIndexIndex(int bufferSize) {
		assert bufferSize > 0 : "bufferSize muss grer 0 sein.";
		bufferSize = bufferSize <= 0 ? 1 : bufferSize;
		rbInit(bufferSize);
	}


	// //////////////////////////// Methoden auf den Current*-Werten ////////////////////////////
	/**
	 * Aktualisiert den Max-Wert des aktuellsten Eintrages. Der Min-Wert und die ContainerID bleiben unverndert
	 *
	 * @param newDIMax neuer Wert fr DatenIndexIndex-Max
	 *
	 * @throws IndexException Wenn der neue Wert kleiner als der dazugehrige Min-Wert ist oder wenn keinen currentEntry gibt
	 */
	public void updateMax(long newDIMax) throws IndexException {
		if(currentEntry[CONTID_POS] == NO_VALUE) throw new IndexException("Update, obwohl noch kein Eintrag hinzugefgt wurde");
		if(newDIMax < currentEntry[MIN_POS]) throw new IndexException("neuer DIMax-Wert ist kleiner als DIMin-Wert");

//        System.out.println("updateMax(" + newDIMax + ");");

		currentEntry[MAX_POS] = newDIMax;
	}


	/**
	 * Liefert die aktuelle Container ID. Wenn es noch keine Aktuelle Container ID gibt, wird -1 zurckgeliefert
	 *
	 * @return die aktuelle Container ID oder -1, wenn es keinen aktuellen Entry gibt.
	 */
	public long getCurrentCID() {
		return currentEntry[CONTID_POS];
	}


	/**
	 * Liefert vom aktuellen Eintrag den Min-Wert des DatenIndex
	 *
	 * @return Min-Wert des DatenIndex des aktuellen Eintrags
	 */
	public long getCurrentDIMin() {
		return currentEntry[MIN_POS];
	}


	/**
	 * Liefert vom aktuellen Eintrag den Max-Wert des DatenIndex
	 *
	 * @return Max-Wert des DatenIndex des aktuellen Eintrags
	 */
	public long getCurrentDIMax() {
		return currentEntry[MAX_POS];
	}


	/**
	 * Anzahl der Eintrge, die dieser Index im RAM halten kann
	 *
	 * @return Anzahl der Eintrge
	 */
	public int capacity() {
		return rbCapacity();
	}

	/**
	 * Anzahl der Eintrge, die dieser Index im RAM hlt
	 *
	 * @return Anzahl der Eintrge
	 */
	public int entries() {
		return rbEntries();
	}


	/**
	 * Fgt dem Index einen neuen Eintrag hinzu. Dabei wird der derzeitig aktuelle Eintrag abgeschlossen. Fhrt dies dazu, dass keine weiteren freien Pltze im
	 * Cache vorhanden sind, werden die nderungen auf Platte geschrieben. Wenn der Cache noch vollstndig leer ist, werden die aktuellsten Daten aus der
	 * Index-Datei geladen.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 * @param temp     Puffer, der beim Einlesen/Schreiben der Index-Datei verwendet wird
	 * @param dIMin    DatenIndex Min-Wert
	 * @param dIMax    DatenIndex Max-Wert
	 * @param contId   Container ID
	 *
	 * @throws IndexException Bei Problemem mit der Index-Datei
	 */
	public void add(String basePath, byte[] temp, long dIMin, long dIMax, long contId) throws IndexException {
		checkForAdd(basePath, temp, dIMin, dIMax, contId);
//        System.out.println("add(" + dIMin + "," + dIMax + "," + contId + ")");

		if(rbMustBeSaved()) flush(basePath);

		if(currentEntry[MIN_POS] < 0) {
			if(biggestDIMaxInFile == NO_VALUE && new File(basePath, IDX_FILENAME).exists()) {
				RandomAccessFile raf = ifOpen(basePath);
				rbLoad(raf);
				ifClose(raf);
			}
		}
		else {
			currentEntryToBuffer(basePath, temp);
			if(biggestDIMaxInFile == NO_VALUE) { // Wir wissen nichts ber die Datei
				if(new File(basePath, IDX_FILENAME).exists()) {
					RandomAccessFile raf = ifOpen(basePath);
					try {
						if(raf.length() > 0) {
							int cachedEntries = (rbOffsetOfFirstUnsavedFromSmallest == NO_VALUE) ? rbEntries() : rbOffsetOfFirstUnsavedFromSmallest;
							if(((int)raf.length() - IndexEntryIndex.ENTRY_LEN * cachedEntries) > 0) {
								biggestDIMaxInFile = getMaximaFromFile(raf, temp, 0, (int)raf.length() - IndexEntryIndex.ENTRY_LEN * cachedEntries)[0][0];
							}
						}
					}
					catch(IOException e) {
						throw new IndexException(e);	// Sollte nie auftreten
					}
					finally {
						ifClose(raf);
					}
				}
			}
		}

		currentEntry[MIN_POS] = dIMin;
		currentEntry[MAX_POS] = dIMax;
		currentEntry[CONTID_POS] = contId;
	}

	/**
	 * Prft, ob dem Index ein neuer Eintrag hinzugefgt werden knnte.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 * @param temp     Puffer, der beim Einlesen/Schreiben der Index-Datei verwendet wird
	 * @param dIMin    DatenIndex Min-Wert
	 * @param dIMax    DatenIndex Max-Wert
	 * @param contId   Container ID
	 *
	 * @throws IndexException Bei Problemem mit der Index-Datei
	 */
	public void checkForAdd(String basePath, byte[] temp, long dIMin, long dIMax, long contId) throws IndexException {
		if(dIMin < 0 || dIMax < 0 || contId <= 0 || dIMin > dIMax) {
			throw new CorruptIndexException("DatenIndexIndex: neuer Eintrag ungltig: (" + "dIMin=" + dIMin + ", dIMax=" + dIMax + ", contID=" + contId + ")");
		}
		if(getCurrentCID() >= contId) {
			throw new CorruptIndexException("DatenIndexIndex: ContainerId des neuen Eintrags ungltig: currentCId=" + getCurrentCID());
		}
	}


	/**
	 * Liefert eine Matrix mit DIMin/Max und ContainerIds, deren DTMin/Max-Werte zwischen dTMin und dTMax liegen. Wird diese Methode aufgerufen wenn der Cache noch
	 * leer ist, werden die aktuellsten Daten aus der Datei nachgeladen.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 * @param temp     Puffer, der beim Einlesen/Schreiben der Index-Datei verwendet wird
	 * @param dTMin    Min-Wert der Zeitspanne
	 * @param dTMax    Max-Wert der Zeitspanne
	 *
	 * @return IndexResult-Objekt, das die Index innerhalb der gesuchten Zeitspanne enthlt (DT und DI Werte, keine AT Werte)
	 *
	 * @throws IndexException Bei Problemem mit der IndexDatei
	 */
	public IndexResult getContainerID(String basePath, byte[] temp, long dIMin, long dIMax) throws IndexException {
		IndexResult ir = new IndexResult();
		RandomAccessFile raf = null;

		if(currentEntry[MIN_POS] == NO_VALUE) {
			try {
				raf = ifOpen(basePath);
				rbLoad(raf);
			}
			catch(IndexException e) {
				return ir;
			}
		}

		if(biggestDIMaxInFile == NO_VALUE) {
			if(new File(basePath, IDX_FILENAME).exists()) {
				if(raf == null) raf = ifOpen(basePath);
				try {
					if(raf.length() > 0) {
						int cachedEntries = (rbOffsetOfFirstUnsavedFromSmallest == NO_VALUE) ? rbEntries() : rbOffsetOfFirstUnsavedFromSmallest;
						if(((int)raf.length() - IndexEntryIndex.ENTRY_LEN * cachedEntries) > 0) {
							biggestDIMaxInFile = getMaximaFromFile(raf, temp, 0, (int)raf.length() - IndexEntryIndex.ENTRY_LEN * cachedEntries)[0][0];
						}
//                        int cachedEntries = (rbOffsetOfFirstUnsavedFromSmallest == NO_VALUE) ? rbEntries() : rbElementsInbetween(rbOffsetOfFirstUnsavedFromSmallest, rbSmallestEntry, rbEntries());
//                        long maxima[][] = getMaximaFromFile(raf, temp, 0, (int) raf.length() - IndexEntryIndex.ENTRY_LEN * cachedEntries);
//                        biggestDIMaxInFile = maxima[0][0];
					}
				}
				catch(IOException e) {
					throw new IndexException(e);	// Sollte nie auftreten
				}
			}
		}

//        if (dIMin < biggestDIMaxInFile && currentEntry[CONTID_POS] != NO_VALUE) {
		if(biggestDIMaxInFile != NO_VALUE && dIMin < biggestDIMaxInFile) {
			if(raf == null) raf = ifOpen(basePath);
			linearFileSearch(raf, temp, ir, dIMin, dIMax);
		}
		ifClose(raf);

X:
		for(int i = (rbGetFirstUnsaved() <= 0 || dIMin >= biggestDIMaxInFile) ? 0 : rbGetFirstUnsaved(); i < rbEntries(); i++) {
			if(overlap(rbGetDIMin(i), rbGetDIMax(i), dIMin, dIMax)) {
				for(int j = ir.size() - 1; j >= 0; j--) {
					if(ir.getCID(j) == rbGetCID(i)) continue X;
				}
				ir.add(rbGetCID(i), rbGetDIMin(i), rbGetDIMax(i));
			}
		}

Y:
		if(currentEntry[MIN_POS] != NO_VALUE) {
			if(overlap(currentEntry[MIN_POS], currentEntry[MAX_POS], dIMin, dIMax)) {
				if(ir.size() == 0 || (ir.getDIMin(ir.size() - 1) < currentEntry[MIN_POS] && filePositionReactivatedEntry == NO_VALUE)) {
					ir.add(currentEntry[CONTID_POS], currentEntry[MIN_POS], currentEntry[MAX_POS]);
					break Y;
				}
				for(int j = ir.size() - 1; j >= 0; j--) {
					if(ir.getCID(j) == currentEntry[CONTID_POS]) {
						if(filePositionReactivatedEntry > 0) ir.updateDIMax(j, currentEntry[MAX_POS]);
						break Y;
					}
					if(ir.getDIMin(j) <= currentEntry[MIN_POS]) {
						ir.insert(j, currentEntry[CONTID_POS], currentEntry[MIN_POS], currentEntry[MAX_POS]);
						break Y;
					}
				}
				ir.insert(0, currentEntry[CONTID_POS], currentEntry[MIN_POS], currentEntry[MAX_POS]);
			}
		}
		return ir;
	}


	/**
	 * Liefert den kleinsten DT Min Wert, der jemals in diesem Index gespeichert wird.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 *
	 * @return Kleinsten Datenindex oder -1, falls keiner vorhanden
	 *
	 * @throws IndexException Bei Zugriffsproblemen mit der IndexDatei
	 */
	public long getAbsolutDIdxMin(String basePath) throws IndexException {
		RandomAccessFile raf = null;
		File f = new File(basePath, IDX_FILENAME);

		if(f.length() == 0) {
			if(currentEntry[MIN_POS] == NO_VALUE) {
				// Fall 1.1: Datei leer, Ram leer           => Leere Antwort
				return NO_VALUE;
			}
			else {
				// Fall 1.2: Datei leer, Ram hlt Daten     => Antwort ist ltestes Element im Ram
				if(rbEntries() > 0) {
					return rbGetDIMin(0);
				}
				else {
					return currentEntry[MIN_POS];
				}
			}
		}
		else {
			if(currentEntry[MIN_POS] == NO_VALUE) {
				// Fall 2.1: Datei hlt Daten, Ram leer     => Daten laden. Antwort ist ltestes Element im Ram, wenn
				//                                             alle Daten ins Ram geladen werden konnten. Ansonsten wie 2.2 behandeln
				raf = ifOpen(basePath);
				rbLoad(raf);
				if(f.length() == rbEntries() * rbEntryLength()) {
					ifClose(raf);
					return rbGetDIMin(0);
				}
			}
			//    Fall 2.2: Daten und Ram enthlt Daten     => ltestes Element aus der Datei nehmen
			if(raf == null) raf = ifOpen(basePath);
			byte tmp[] = new byte[rbEntryLength()];
			int bytesRead = -1;
			try {
				raf.seek(0);
				bytesRead = raf.read(tmp, 0, tmp.length);
			}
			catch(IOException e) {
			}
			finally {
				ifClose(raf);
			}

			if(bytesRead == rbEntryLength()) return IndexEntryIndex.getDIMin(tmp, 0);

			return NO_VALUE;
		}
	}

	/**
	 * Schliet den Index und speichert alle genderten Daten auf Platte. Nach dem Aufruf von close() darf die Index-Instanz nicht mehr verwendet werden.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 * @param temp     Puffer, der beim Schreiben der Index-Datei verwendet wird
	 *
	 * @throws IndexException
	 */
	public void closeIndex(String basePath, byte[] temp) throws IndexException {
//    	System.out.println("closeIndex()");
		if(rbMustBeSaved()) flush(basePath);
		if(currentEntry[MIN_POS] != NO_VALUE) currentEntryToBuffer(basePath, temp);
		flush(basePath);
	}


	/**
	 * Gibt den Inhalt des ByteBuffers als String aus, inkl Angaben ber Kapazitt, Anzahl der Eintrge und des CurrentEntries.
	 *
	 * @return ByteBuffer als String
	 */
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("capacity: " + rbCapacity() + ", entries: " + rbEntries());
		sb.append(", rbSmallestEntry: " + rbSmallestEntry);
		sb.append(", filePositionReactivatedEntry: " + filePositionReactivatedEntry);
		sb.append(
				", current: (" + currentEntry[MIN_POS] + "," + currentEntry[MAX_POS] + "," + currentEntry[CONTID_POS] + ") ringbuffer: "
		);
		return sb.toString() + rbToString();
	}


	// ///////////////////////////// Private Helpermethoden /////////////////////////////////
	private void currentEntryToBuffer(String basePath, byte[] temp) throws IndexException {
		if(filePositionReactivatedEntry != NO_VALUE) {
			updateCurrentEntryInFile(basePath);
		}
		else if(currentEntry[MIN_POS] >= rbGetDIMin(0)) {
			rbAddCurrentEntry();
		}
		else {
			insertIntoFile(basePath, temp);
		}
	}


	private void flush(String basePath) throws IndexException {
		if(rbHasUnsavedData()) {
			try {
				if(rbCacheChanged()) {
					RandomAccessFile raf = ifOpen(basePath);
					try {
						long entry[] = rbGetFileAnchor();
						byte buffer[] = new byte[(raf.length() > rbCapacityInBytes()) ? rbCapacityInBytes() : (int)raf.length()];
						raf.seek(raf.length() - buffer.length);
						raf.read(buffer);
						int idx = locateEntry(buffer, entry);
						if(idx > 0) raf.setLength(raf.length() - (idx * IndexEntryIndex.ENTRY_LEN));
						rbAppendToFile(raf);
					}
					finally {
						ifClose(raf);
					}
				}
				else
				{
					final FileOutputStream os = new FileOutputStream(new File(basePath, IDX_FILENAME), true);
					try {
						osAppendToFile(os);
					}
					finally {
						os.close();
					}
				}
				rbOffsetOfFirstUnsavedFromSmallest = NO_VALUE;
			}
			catch(Exception e) {
				throw new IndexException(ERROR_ACCESS + e.getMessage(), e);
			}
		}
	}


	private void insertIntoFile(String basePath, byte temp[]) throws IndexException {
		RandomAccessFile raf = null;
		try {
			raf = ifOpen(basePath);
			byte entry[] = new byte[rbEntryLength()];
			IndexEntryIndex.setEntry(entry, 0, currentEntry);

			int start = 0;
			int end = (int)raf.length();

			if(end == 0) {
				raf.write(entry); // Leere Datei, kein umkopieren
			}
			else {
				int position = (int)binaryFileSearch(raf, temp, currentEntry[MIN_POS]);

				while(start >= 0) {
					start = end - temp.length;
					if(start <= position) {
						raf.seek(position);
						int bytesRead = raf.read(temp, 0, (end - position));
						raf.seek(position);
						raf.write(entry);
						if(bytesRead > 0) raf.write(temp, 0, bytesRead);
						break;
					}
					else {
						raf.seek(start);
						int bytesRead = raf.read(temp);
						raf.seek(start + entry.length);
						raf.write(temp, 0, bytesRead);
					}
					end = start;
				}
			}
			if(currentEntry[MAX_POS] > biggestDIMaxInFile) {
				biggestDIMaxInFile = currentEntry[MAX_POS];
			}
		}
		catch(Exception e) {
			throw new CorruptIndexException("Probleme beim Umkopieren der Indexdatei. " + "Sie ist nun sehr wahrscheinlich beschdigt: " + e.toString(), e);
		}
		finally {
			ifClose(raf);
		}
	}


	/**
	 * Ermittelt mit einer binren Suche die Position, an der der Key1 untergebracht werden kann. Wenn es schon mehrere gleiche Key1 gibt, wird ein beliebiger
	 * davon zurck gegeben.
	 *
	 * @param raf  Eine Referenz auf die Datei
	 * @param temp Ein Temp Array, zum Einlesen der Datei. Je grer umso besser
	 * @param key  der gesuchte Key1
	 *
	 * @return die bytePosition an der der Key eingefgt werden muss
	 *
	 * @throws IndexException
	 * @throws IOException
	 */
	private long binaryFileSearch(RandomAccessFile raf, byte[] temp, long key) throws IndexException, IOException {
		assert temp.length >= rbEntryLength() : "Puffer kleiner als Lnge eines Eintrags";
		if(temp.length < rbEntryLength()) temp = new byte[(1024 / rbEntryLength()) * rbEntryLength()];

		int useableArea = (temp.length / rbEntryLength()) * rbEntryLength();
		int numberOfChunks = ((int)raf.length() + useableArea - 1 )/ useableArea;

		// Denn richtigen Abschnitt in der Datei finden
		int fMid = 0, fLow = 0, bytesRead = -1;
		int fHigh = numberOfChunks - 1;
		while(fLow <= fHigh) {
			fMid = (fLow + fHigh) >> 1;
			raf.seek(fMid * useableArea);
			bytesRead = raf.read(temp, 0, useableArea);

			int cmp = containsKey(temp, bytesRead / IndexEntryIndex.ENTRY_LEN, key);
			if(cmp < 0) {
				fHigh = fMid - 1;
			}
			else if(cmp > 0) {
				fLow = fMid + 1;
			}
			else {
				break;
			}
		}

		// und jetzt den richtigen Eintrag im eingelesenen Chunk suchen
		int low = 0, mid = 0;
		int high = bytesRead / IndexEntryIndex.ENTRY_LEN - 1;
		int max = high;
		while(low <= high) {
			mid = (low + high) >> 1;
			long midVal = IndexEntryIndex.getDIMin(temp, mid);

			if(midVal < key) {
				low = mid + 1;
			}
			else if(midVal > key) {
				high = mid - 1;
			}
			else {
				while(midVal == key){
					mid++;
					if(mid>max) break;
					midVal = IndexEntryTime.getDTMin(temp, mid);
				}
				return mid * IndexEntryIndex.ENTRY_LEN + fMid * useableArea;
			}
		}
		return low * IndexEntryIndex.ENTRY_LEN + fMid * useableArea;
	}


	private void linearFileSearch(RandomAccessFile raf, byte[] temp, IndexResult ir, long min, long max) throws IndexException {
		assert temp.length >= rbEntryLength() : "Puffer kleiner als Lnge eines Eintrags";

		if(temp.length < rbEntryLength()) temp = new byte[(1024 / rbEntryLength()) * rbEntryLength()];

		try {
			int useableArea = (temp.length / rbEntryLength()) * rbEntryLength();
			int chunks = ((int)raf.length() / useableArea);

			for(int i = 0; i <= chunks; i++) {
				raf.seek(i * useableArea);
				int bytesRead = raf.read(temp, 0, useableArea);

				int entries = bytesRead / IndexEntryIndex.ENTRY_LEN;
				for(int j = 0; j < entries; j++) {
					if(overlap(IndexEntryIndex.getDIMin(temp, j), IndexEntryIndex.getDIMax(temp, j), min, max)) {
						ir.add(IndexEntryIndex.getCID(temp, j), IndexEntryIndex.getDIMin(temp, j), IndexEntryIndex.getDIMax(temp, j));
					}
				}
			}
		}
		catch(Exception e) {
			throw new IndexException(ERROR_ACCESS + e.getMessage(), e);
		}
	}


	private static int locateEntry(byte buffer[], long entry[]) {
		int elements = buffer.length / IndexEntryIndex.ENTRY_LEN;
		for(int i = 0; i < elements; i++) {
			final long diMin = IndexEntryIndex.getDIMin(buffer, i);
			final long currentDiMin = entry[MIN_POS];
			if(diMin > currentDiMin || (diMin == currentDiMin && IndexEntryIndex.getCID(buffer, i) == entry[CONTID_POS])) {
				return elements - i;
			}
		}
		return -1;
	}


	private void updateCurrentEntryInFile(String basePath) throws IndexException {
		assert filePositionReactivatedEntry != NO_VALUE : "updateCurrentEntry(), obwohl es keinen wiederaufgenommen Eintrag gibt";

		RandomAccessFile raf = null;
		try {
			raf = ifOpen(basePath);
			raf.seek(filePositionReactivatedEntry);
			byte temp[] = new byte[rbEntryLength()];
			IndexEntryIndex.setEntry(temp, 0, currentEntry);
			raf.write(temp);
			filePositionReactivatedEntry = NO_VALUE;
			rbOffsetOfFirstUnsavedFromSmallest = NO_VALUE;
			rbLoad(raf);
		}
		catch(Exception e) {
			throw new IndexException(ERROR_ACCESS + e.getMessage(), e);
		}
		finally {
			ifClose(raf);
		}
	}


	/**
	 * @param raf       Random Access File
	 * @param temp      Temp. Buffer
	 * @param fileStart Erstes Byte, ab den gelesen werden soll
	 * @param fileEnd   Letztes Byte, bis zu dem gelesen werden soll
	 *
	 * @return Matrix:  [0][0] = Grter DTMax-Wert [1][0] = erstes Byte des Eintrags mit der grten CID [2][]  = der Maximaeintrag selbst
	 *
	 * @throws IOException
	 */
	private long[][] getMaximaFromFile(RandomAccessFile raf, byte[] temp, int fileStart, int fileEnd) throws IOException {
		assert raf != null : "Kein RandomAccessfile";
		assert temp != null && temp.length >= rbEntryLength() : "temp nicht vorhanden oder nicht gro genug";
		assert fileStart <= fileEnd : "fileStart muss kleienr als FileEnd sein";

		long maxima[][] = {
				{NO_VALUE}, {NO_VALUE}, {NO_VALUE, NO_VALUE, NO_VALUE}
		};

		if(raf.length() <= 0) return maxima;

		// byteBuffer in benutzbare Bereiche einteilen
		int useableArea = ((int)temp.length / rbEntryLength()) * rbEntryLength();
		int chunks = ((int)(fileEnd - fileStart) / useableArea) + 1;

		for(int i = 0; i < chunks; i++) {
			raf.seek(i * fileStart);
			int bytesRead = raf.read(temp, 0, useableArea);
			int entries = bytesRead / rbEntryLength();
			for(int j = 0; j < entries && (i * useableArea + fileStart + j * rbEntryLength()) < fileEnd; j++) {
				if(IndexEntryIndex.getDIMax(temp, j) > maxima[2][MAX_POS]) // grter DIMax
				{
					maxima[0][0] = IndexEntryIndex.getDIMax(temp, j);
				}

				if(IndexEntryIndex.getCID(temp, j) > maxima[2][CONTID_POS]) { // Eintrag mit grter CID
					maxima[2] = IndexEntryIndex.getEntry(temp, j);
					maxima[1][0] = (i * useableArea) + fileStart + j * rbEntryLength();
				}
			}
		}
		return maxima;
	}


	private int containsKey(byte[] array, int cnt, long key) {
		return key < IndexEntryIndex.getDIMin(array, 0) ? -1 : (key >= IndexEntryIndex.getDIMin(array, cnt - 1) ? 1 : 0);
	}


	private boolean overlap(long firstMin, long firstMax, long secondMin, long secondMax) {
		return (secondMin <= firstMax && secondMin >= firstMin) || (secondMax <= firstMax && secondMax >= firstMin)
		       || (firstMin <= secondMax && firstMin >= secondMin) || (firstMax <= secondMax && firstMax >= secondMin);
	}


	// ///////////////////////////////////////////////////////////////////////
	// IndexFile
	private RandomAccessFile ifOpen(String basePath) throws IndexException {
		RandomAccessFile raf = null;

		try {
			raf = new RandomAccessFile(new File(basePath, IDX_FILENAME), "rw");
			return raf;
		}
		catch(IOException e) {
			throw new IndexException(ERROR_OPEN + e.getMessage(), e);
		}
	}


	private void ifClose(RandomAccessFile raf) throws IndexException {
		if(raf != null) {
			try {
				raf.close();
			}
			catch(IOException e) {
				throw new IndexException(ERROR_CLOSE + e.getMessage(), e);
			}
		}
	}

	// ///////////////////////////////////////////////////////////////////////
	// Ab hier ist alles, was den RingPuffer betrifft. Oberhalb dieses
	// Kommentars drfen nur die rp* Methoden verwendet werden, nicht
	// aber die Felder
	private byte rbBuffer[] = null;

	private int rbSmallestEntry = 0;

	private int rbBiggestEntry = NO_VALUE;

	private int rbOffsetOfFirstUnsavedFromSmallest = NO_VALUE;

	private boolean rbFileupdateNecessary = false;


	private void rbInit(int entries) {
		rbBuffer = new byte[entries * IndexEntryIndex.ENTRY_LEN];
	}


	private void rbLoad(RandomAccessFile raf) throws IndexException {
		try {
			raf.seek(raf.length() > rbCapacityInBytes() ? raf.length() - rbCapacityInBytes() : 0);
			int numRead = raf.read(rbBuffer);
			if(numRead > 0 && numRead % rbEntryLength() != 0) throw new IndexException("Probleme beim Einlesen der Indexe. Ein Eintrag war unvollstndig");

			if(numRead > 0) {
				rbSmallestEntry = 0;
				rbBiggestEntry = numRead / rbEntryLength() - 1;
			}
			else {
				rbSmallestEntry = 0;
				rbBiggestEntry = -1;
			}
		}
		catch(Exception e) {
			throw new IndexException(ERROR_ACCESS + e.getMessage(), e);
		}
	}


	/**
	 * Verndert die Gre des RingBuffers. Beim Verkleinern werden nicht gespeicherte Eintrge auf Platte geschrieben
	 *
	 * @param newSize
	 * @param basePath
	 *
	 * @throws IndexException
	 */
	public void resize(int newSize, String basePath) throws IndexException {
		assert newSize > 0 : "Neue Gre kleiner/gleich 0";
		newSize = newSize > 0 ? newSize : 1;

		if(rbCapacity() == newSize)             // Kein Umkopieren ntig, da gleiche Gre
		{
			return;
		}
		if(rbEntries() == 0) {                  //  Keine Daten -> nur neuen Buffer anlegen
			rbBuffer = new byte[newSize * IndexEntryIndex.ENTRY_LEN];
			return;
		}
		if(newSize > rbCapacity()) {                 // vergrern
			if(rbSmallestEntry <= rbBiggestEntry) { // einfaches umkopieren reicht aus
				byte oldBuffer[] = rbBuffer;
				int start = rbBytePos(0);
				int length = rbBytePos(rbEntries() - 1) - start + IndexEntryIndex.ENTRY_LEN;
				rbBuffer = new byte[newSize * IndexEntryIndex.ENTRY_LEN];
				System.arraycopy(oldBuffer, start, rbBuffer, start, length);
			}
			else {                                // Daten berlappen Grenze. Neu anordnen und Zeiger anpassen
				byte oldBuffer[] = rbBuffer;
				int num = rbEntries();
				int start = rbBytePos(0);
				int length = rbCapacityInBytes() - start;
				rbBuffer = new byte[newSize * IndexEntryIndex.ENTRY_LEN];
				System.arraycopy(oldBuffer, start, rbBuffer, 0, length);
				start = length;
				length = rbBiggestEntry * IndexEntryIndex.ENTRY_LEN + IndexEntryIndex.ENTRY_LEN;
				System.arraycopy(oldBuffer, 0, rbBuffer, start, length);
				rbBiggestEntry = num - 1;
				rbSmallestEntry = 0;
			}
		}
		else {                                    // Verkleinern
			// Speichern, wenn es ungesicherte Daten gibt und diese mehr Platz belegen, als im neuen Puffer zur Verfgung stehen
			if(rbHasUnsavedData() && rbEntries() - rbOffsetOfFirstUnsavedFromSmallest > newSize) flush(basePath);
			int firstToCopy = (newSize > rbEntries()) ? rbSmallestEntry : (rbEntries() + rbBiggestEntry + 1 - newSize) % rbEntries();

			if(firstToCopy < rbBiggestEntry) {
				byte oldBuffer[] = rbBuffer;
				int start = firstToCopy * IndexEntryIndex.ENTRY_LEN;
				int length = rbBytePos(rbEntries() - 1) - start + IndexEntryIndex.ENTRY_LEN;
				rbBuffer = new byte[newSize * IndexEntryIndex.ENTRY_LEN];
				System.arraycopy(oldBuffer, start, rbBuffer, 0, length);
				if(rbOffsetOfFirstUnsavedFromSmallest > 0) {
					rbOffsetOfFirstUnsavedFromSmallest = (length / IndexEntryIndex.ENTRY_LEN - 1) - (rbBiggestEntry - firstToCopy);
				}
				rbSmallestEntry = 0;
				rbBiggestEntry = length / IndexEntryIndex.ENTRY_LEN - 1;
			}
			else {
				byte oldBuffer[] = rbBuffer;
				int start = firstToCopy * IndexEntryIndex.ENTRY_LEN;
				int length = rbCapacityInBytes() - start;
				rbBuffer = new byte[newSize * IndexEntryIndex.ENTRY_LEN];
				System.arraycopy(oldBuffer, start, rbBuffer, 0, length);
				start = length;
				length = rbBiggestEntry * IndexEntryIndex.ENTRY_LEN + IndexEntryIndex.ENTRY_LEN;
				System.arraycopy(oldBuffer, 0, rbBuffer, start, length);
				if(rbOffsetOfFirstUnsavedFromSmallest > 0) {
					rbOffsetOfFirstUnsavedFromSmallest = rbOffsetOfFirstUnsavedFromSmallest - ((oldBuffer.length / IndexEntryIndex.ENTRY_LEN) - newSize);
				}
				rbBiggestEntry = (start + length) / IndexEntryIndex.ENTRY_LEN - 1;
				rbSmallestEntry = 0;
			}
		}
	}


	private int rbElementsInbetween(int start, int stop, int ringbufferSize) {
		if(start <= stop) return stop - start;
		return (ringbufferSize + (stop - start)) + 1;
	}


	private int rbEntries() {
		if(rbBiggestEntry == -1) {
			return 0;
		}
		else if(rbSmallestEntry > rbBiggestEntry) {
			return rbCapacity() - rbSmallestEntry + rbBiggestEntry + 1;
		}
		else {
			return rbBiggestEntry - rbSmallestEntry + 1;
		}
	}


	private int rbBytePos(int pos) {
		if(rbSmallestEntry + pos < rbCapacity()) {
			return (rbSmallestEntry + pos) * IndexEntryIndex.ENTRY_LEN;
		}
		else {
			return (pos - rbCapacity() + rbSmallestEntry) * IndexEntryIndex.ENTRY_LEN;
		}
	}


	private int rbCapacity() {
		return rbBuffer.length / IndexEntryIndex.ENTRY_LEN;
	}


	private int rbCapacityInBytes() {
		return rbBuffer.length;
	}


	private int rbEntryLength() {
		return IndexEntryIndex.ENTRY_LEN;
	}


	private boolean rbMustBeSaved() {
		return (rbOffsetOfFirstUnsavedFromSmallest == 0 && rbEntries() == rbCapacity());
	}


	private void rbAddCurrentEntry() {
		if(currentEntry[MIN_POS] >= rbGetDIMin(rbEntries() - 1)) {
			rbAppendCurrentEntry();
		}
		else {
			rbInsertElement();
		}
	}


	private void rbAppendCurrentEntry() {
		if(rbEntries() > 0 && IndexEntryIndex.getDIMax(rbBuffer, rbStep(rbBiggestEntry)) > biggestDIMaxInFile) {
			biggestDIMaxInFile = IndexEntryIndex.getDIMax(rbBuffer, rbStep(rbBiggestEntry));
		}

		IndexEntryIndex.setEntry(rbBuffer, rbStep(rbBiggestEntry), currentEntry);
		if(rbEntries() == rbCapacity() && rbStep(rbBiggestEntry) == rbSmallestEntry) {
			rbSmallestEntry = rbStep(rbSmallestEntry);
			if(rbOffsetOfFirstUnsavedFromSmallest >= 0) rbOffsetOfFirstUnsavedFromSmallest = rbStepBack(rbOffsetOfFirstUnsavedFromSmallest);
		}

		rbBiggestEntry = rbStep(rbBiggestEntry);
		if(rbOffsetOfFirstUnsavedFromSmallest < 0) rbOffsetOfFirstUnsavedFromSmallest = rbEntries() - 1;
	}


	private void rbInsertElement() {
		if(rbCapacity() == rbEntries()) {
			if(biggestDIMaxInFile != NO_VALUE && IndexEntryIndex.getDIMax(rbBuffer, rbSmallestEntry) > biggestDIMaxInFile) {
				biggestDIMaxInFile = IndexEntryIndex.getDIMax(rbBuffer, rbSmallestEntry);
			}
		}
		int position = rbFindDIMin(currentEntry[MIN_POS]);
		for(int i = rbEntries() - 1; i >= position; i--) {
			rbMoveEntryRight(i);
		}
		IndexEntryIndex.setEntry(rbBuffer, (rbSmallestEntry + position) % rbCapacity(), currentEntry);

		if(rbEntries() == rbCapacity() && rbStep(rbBiggestEntry) == rbSmallestEntry) {
			rbSmallestEntry = rbStep(rbSmallestEntry);
			position = rbStepBack(position);
		}
		rbBiggestEntry = rbStep(rbBiggestEntry);

		if(rbOffsetOfFirstUnsavedFromSmallest == NO_VALUE || (rbOffsetOfFirstUnsavedFromSmallest != NO_VALUE
		                                                      && position < rbOffsetOfFirstUnsavedFromSmallest)) {
			rbOffsetOfFirstUnsavedFromSmallest = position;
			rbFileupdateNecessary = true;
		}
		else if(rbEntries() == rbCapacity() && rbOffsetOfFirstUnsavedFromSmallest > 0) {
			rbOffsetOfFirstUnsavedFromSmallest = rbStepBack(rbOffsetOfFirstUnsavedFromSmallest);
		}
	}


	private void rbExtractMaxima() {
		long tmpEntry[] = new long[currentEntry.length];
		System.arraycopy(currentEntry, 0, tmpEntry, 0, tmpEntry.length);
		int tmpIndex = -1;
		for(int i = rbEntries() - 1; i >= 0; i--) {
			if(rbGetCID(i) >= tmpEntry[CONTID_POS]) {
				tmpEntry = rbGetEntry(i);
				tmpIndex = i;
			}
		}
		currentEntry = tmpEntry;

		for(int i = tmpIndex + 1; i < rbCapacity(); i++) {
			rbMoveEntryLeft(i);
		}

		if(rbEntries() == 1) {
			rbBiggestEntry = -1; // also leerer ringbuffer
		}
		else {
			rbBiggestEntry = rbStepBack(rbBiggestEntry);
		}
	}


	private void rbMoveEntryRight(int index) {
		System.arraycopy(rbBuffer, rbBytePos(index), rbBuffer, rbBytePos(rbStep(index)), IndexEntryIndex.ENTRY_LEN);
	}


	private void rbMoveEntryLeft(int index) {
		System.arraycopy(rbBuffer, rbBytePos(index), rbBuffer, rbBytePos(rbStepBack(index)), IndexEntryIndex.ENTRY_LEN);
	}


	private int rbFindDIMin(long key) {
		int low = 0, high = rbEntries() - 1, mid;
		int max = high;
		while(low <= high) {
			mid = (low + high) / 2;
			long midVal = rbGetDIMin(mid);
			if(midVal > key) {
				high = mid - 1;
			}
			else if(midVal < key) {
				low = mid + 1;
			}
			else {
				while(midVal == key){
					mid++;
					if(mid>max) break;
					midVal = rbGetDIMin(mid);
				}
				return mid;
			}
		}
		return low;
	}


	private long rbGetDIMin(int index) {
		if(index < 0) {
			return -1;
		}
		else {
			return IndexEntryIndex.getDIMin(rbBuffer, (rbSmallestEntry + index) % rbCapacity());
		}
	}


	private long rbGetDIMax(int index) {
		if(index < 0) {
			return -1;
		}
		else {
			return IndexEntryIndex.getDIMax(rbBuffer, (rbSmallestEntry + index) % rbCapacity());
		}
	}


	private long rbGetCID(int index) {
		if(index < 0) {
			return -1;
		}
		else {
			return IndexEntryIndex.getCID(rbBuffer, (rbSmallestEntry + index) % rbCapacity());
		}
	}

	private long[] rbGetEntry(int index) {
		return new long[]{rbGetDIMin(index), rbGetDIMax(index), rbGetCID(index)};
	}


	private int rbStep(int index) {
		return (index + 1) % rbCapacity();
	}


	private int rbStepBack(int index) {
		return (index + rbCapacity() - 1) % rbCapacity();
	}


	private void rbAppendToFile(RandomAccessFile raf) throws Exception {
		int first = (rbSmallestEntry + rbOffsetOfFirstUnsavedFromSmallest) % rbCapacity();
		int last = rbBiggestEntry;
		raf.seek(raf.length());
		if(first > last) {
			byte[] tmpArray = new byte[rbBuffer.length - (first * rbEntryLength()) + last * rbEntryLength() + rbEntryLength()];
			System.arraycopy(rbBuffer, first * rbEntryLength(), tmpArray, 0, rbBuffer.length - (first * rbEntryLength()));
			System.arraycopy(rbBuffer, 0, tmpArray, rbBuffer.length - (first * rbEntryLength()), last * rbEntryLength() + rbEntryLength());
			raf.write(tmpArray);
		}
		else {
			// In der ursprnglichen Version von bea wurde die Anzahl zu schreibender Bytes falsch berechnet, was zu fehlenden und fehlerhaften Indexeintrgen
			// fhren konnte
			raf.write(rbBuffer, first * rbEntryLength(), (last - first) % rbCapacity() * rbEntryLength() + rbEntryLength());
		}
	}

	private void osAppendToFile(OutputStream os) throws Exception {
		int first = (rbSmallestEntry + rbOffsetOfFirstUnsavedFromSmallest) % rbCapacity();
		int last = rbBiggestEntry;
		if(first > last) {
			os.write(rbBuffer, first * rbEntryLength(), rbBuffer.length - (first * rbEntryLength()));
			os.write(rbBuffer, 0, last * rbEntryLength() + rbEntryLength());
		}
		else {
			// In der ursprnglichen Version von bea wurde die Anzahl zu schreibender Bytes falsch berechnet, was zu fehlenden und fehlerhaften Indexeintrgen
			// fhren konnte
			os.write(rbBuffer, first * rbEntryLength(), (last - first) % rbCapacity() * rbEntryLength() + rbEntryLength());
		}
	}

	private long[] rbGetFileAnchor() {
		if(rbOffsetOfFirstUnsavedFromSmallest >= 0) {
			return IndexEntryIndex.getEntry(rbBuffer, (rbSmallestEntry + rbOffsetOfFirstUnsavedFromSmallest) % rbCapacity());
		}
		else {
			return new long[]{-1, -1, -1, -1, -1};
		}
	}


	private boolean rbHasUnsavedData() {
		return rbOffsetOfFirstUnsavedFromSmallest >= 0;
	}


	private boolean rbCacheChanged() {
		return rbFileupdateNecessary;
	}


	private int rbGetFirstUnsaved() {
		return rbOffsetOfFirstUnsavedFromSmallest;
	}


	private String rbToString() {
		StringBuilder sb = new StringBuilder();
		sb.append("{");
		for(int i = 0; i < rbCapacity(); i++) {
			if(rbSmallestEntry == i) sb.append("[");
			sb.append(IndexEntryIndex.getDIMin(rbBuffer, i) + "," + IndexEntryIndex.getDIMax(rbBuffer, i) + ",");
			sb.append(IndexEntryIndex.getCID(rbBuffer, i));
			if(rbBiggestEntry == i) sb.append("]");
			sb.append("; ");
		}
		sb.delete(sb.length() - 2, sb.length());
		sb.append("}");
		return sb.toString();
	}
}
