/*
 * Copyright 2009 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.repair;

import de.bsvrz.ars.ars.persistence.ContainerDataIterator;
import de.bsvrz.ars.ars.persistence.ContainerFile;
import de.bsvrz.ars.ars.persistence.ContainerFileReader;
import de.bsvrz.ars.ars.persistence.ContainerHdr;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.index.ATimeMonotoneIndex;
import de.bsvrz.ars.ars.persistence.index.DIdxATimeMonotoneIndex;
import de.bsvrz.ars.ars.persistence.index.DataIndexIndex;
import de.bsvrz.ars.ars.persistence.index.DataTimeIndex;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;

import java.io.BufferedInputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.*;
import java.util.regex.Pattern;

/**
 * Hilfsprogramm, mit dem es mglich ist, Container-Dateien des Archivsystems, in denen bestimmte Defekte vorliegen zu identifizieren und zu korrigieren. Das
 * Programm durchsucht in einem vorgegebenen Verzeichnis und auf Wunsch auch rekursiv in allen Unterverzeichnisse nach Container-Dateien und fhrt Prfungen auf
 * bestimmte Defekte durch, die durch Fehler in lteren Versionen des Archivsystems ausgelst wurden. Defekte Containerdateien werden durch eine entsprechende
 * Ausgabe identifiziert und auf Wunsch wenn mglich auch korrigiert.
 * <p/>
 * Das Programm untersttzt folgende Aufrufparameter: -verzeichnis=... -rekursiv=ja|nein -reparieren=nein|ja -index=nein|ja -container=nein|ja
 * <p/>
 * Mit <code>-verzeichnis=...</code> kann ein beliebiges Verzeichnis angegeben werden, in dem Containerdateien gesucht werden. Defaultwert ist das aktuelle
 * Arbeitsverzeichnis in dem das Programm gestartet wurde.
 * <p/>
 * Mit <code>-rekursiv=ja|nein</code> kann spezifiziert werden, ob auch Unterverzeichnisse des Ausgangsverzeichnisses rekursiv durchsucht werden sollen oder
 * nicht. Ohne Angabe dieses Arguments werden auch Unterverzeichnisse rekursiv durchsucht.
 * <p/>
 * Mit <code>-reparieren=nein|ja</code> kann spezifiziert werden, ob fehlerhafte Dateien nur identifiziert werden sollen oder ob sie auch repariert werden
 * sollen.
 * <p/>
 * Mit <code>-index=ja|nein</code> kann spezifiziert werden, ob Indexdateien berprft werden
 * sollen.
 * <p/>
 * Mit <code>-container=ja|nein</code> kann spezifiziert werden, ob Containerdateien berprft werden
 * sollen.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 9587 $
 */
public class ContainerRescue {

	private File _startDirectory;

	private boolean _recursiveScan;

	private boolean _fixDefects;

	private boolean _checkContainerFiles;

	private boolean _checkIndexFiles;



	private final FilenameFilter _containerFilenameFilter;

	private Pattern _containerFilenamePattern;

	private final FilenameFilter _indexFilenameFilter;

	private Pattern _indexFilenamePattern;

	private FileFilter _directoryFileFilter;

	private List<Defect> _cantFix;

	private List<Defect> _oneByteShortDefects;

	private long _validContainerFileCount = 0;

	private long _checkedContainerFileCount = 0;

	private long _defectContainerFileCount = 0;

	private long _validIndexFileCount = 0;

	private long _checkedIndexFileCount = 0;

	private long _defectIndexFileCount = 0;

	private List<Defect> _badRangeEndDefects;

	private List<Defect> _badIndexFiles;

	private long _correctedDefectCount = 0;

	public static void main(String[] args) {
		try {
			final ArgumentList arguments = new ArgumentList(args);
			final ContainerRescue containerRescue = new ContainerRescue(arguments);
			containerRescue.start();
		}
		catch(Exception e) {
			e.printStackTrace();
			System.err.println("Ein unerwarteter Fehler ist whrend der Programmausfhrung aufgetreten: " + e);
		}
	}

	public ContainerRescue(final ArgumentList arguments) {
		_startDirectory = arguments.fetchArgument("-verzeichnis=.").asDirectory();
		_recursiveScan = arguments.fetchArgument("-rekursiv=ja").booleanValue();
		_checkContainerFiles = arguments.fetchArgument("-container=ja").booleanValue();
		_checkIndexFiles = arguments.fetchArgument("-index=ja").booleanValue();
		_fixDefects = arguments.fetchArgument("-reparieren=nein").booleanValue();
		arguments.ensureAllArgumentsUsed();
		_cantFix = new ArrayList<Defect>();
		_oneByteShortDefects = new ArrayList<Defect>();
		_badRangeEndDefects = new ArrayList<Defect>();
		_badIndexFiles = new ArrayList<Defect>();
		_containerFilenamePattern = Pattern.compile("dc[0-9]*\\.dat");
		_containerFilenameFilter = new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return _containerFilenamePattern.matcher(name).matches();
			}
		};
		_indexFilenamePattern = Pattern.compile("_.+\\.idx");
		_indexFilenameFilter = new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return _indexFilenamePattern.matcher(name).matches();
			}
		};
		_directoryFileFilter = new FileFilter() {
			public boolean accept(File file) {
				return file.isDirectory();
			}
		};
	}

	private void start() {
		System.out.println(
				"Aufrufparameter: -verzeichnis=" + _startDirectory + " -rekursiv=" + (_recursiveScan ? "ja" : "nein") + " -container="
				+ (_checkContainerFiles ? "ja" : "nein") + " -index=" + (_checkIndexFiles ? "ja" : "nein") + " -reparieren=" + (_fixDefects ? "ja" : "nein")
		);
		if(!_checkContainerFiles && !_checkIndexFiles){
			System.out.println("########## Bei -container=nein und -index=nein wird nichts geprft.");
			return;
		}

		System.out.println(
				"########## Das Verzeichnis " + _startDirectory.getAbsolutePath()
				+ (_recursiveScan ? " und rekursiv alle enthaltenen Unterverzeichnisse werden" : " wird") + " nach defekten "
				+ (_checkContainerFiles ? "Container" : "") + ((_checkContainerFiles && _checkIndexFiles) ? "- und " : "") + (_checkIndexFiles ? "Index" : "")
				+ "-Dateien untersucht. " + "Defekte Dateien werden ausgegeben " + (_fixDefects ? "und" : "aber nicht") + " repariert."
		);
		long t0 = System.currentTimeMillis();
		scan(_startDirectory);
		long t1 = System.currentTimeMillis();
		long scanTime = t1 - t0;
		System.out.println("########## Zusammenfassung der Prfung:");
		System.out.println("Dauer der Prfung:                " + String.format("%.3f Sekunden", scanTime / 1000.0));
		System.out.println("Anzahl geprfter Container:       " + _checkedContainerFileCount);
		System.out.println("Anzahl Container ohne Defekte:    " + _validContainerFileCount);
		System.out.println("Anzahl Container mit Defekten:    " + _defectContainerFileCount);
		System.out.println("Anzahl geprfter Indexdateien:    " + _checkedIndexFileCount);
		System.out.println("Anzahl Indexdateien ohne Defekte: " + _validIndexFileCount);
		System.out.println("Anzahl Indexdateien mit Defekten: " + _defectIndexFileCount);
		final long fixableDefectsCount = (_oneByteShortDefects.size() + _badRangeEndDefects.size() + _badIndexFiles.size());
		System.out.println("Anzahl korrigierbare Defekte :    " + fixableDefectsCount);
		if(fixableDefectsCount > 0) {
			if(_fixDefects) {
				if(_oneByteShortDefects.size() > 0) {
					System.out.println("########## Korrektur der gelschten Container, die ein Byte zu kurz sind. Anzahl: " + _oneByteShortDefects.size());
					final List<Defect> oneByteShortDefects = _oneByteShortDefects;
					_oneByteShortDefects = new ArrayList<Defect>();
					for(Defect oneByteShortDefect : oneByteShortDefects) {
						fixOneByteShortDefect(oneByteShortDefect);
					}
					System.out.println("  ########## Erneute Prfung der korrigierten Container:");
					for(Defect oneByteShortDefect : oneByteShortDefects) {
						checkContainerFile(oneByteShortDefect.getFile());
					}
					final long correctedDefectCount = (oneByteShortDefects.size() - _oneByteShortDefects.size());
					System.out.println("  ########## Anzahl erfolgreich korrigierter Defekte: " + correctedDefectCount);
					_correctedDefectCount += correctedDefectCount;
				}
				if(_badRangeEndDefects.size() > 0) {
					System.out.println(
							"########## Korrektur der gelschten Container, die eine falsche Endangabe eines Datensatzindexbereichs enthalten. Anzahl: "
							+ _badRangeEndDefects.size()
					);
					final List<Defect> badRangeEndDefects = _badRangeEndDefects;
					_badRangeEndDefects = new ArrayList<Defect>();
					for(Defect badRangeEndDefect : badRangeEndDefects) {
						fixBadRangeEndDefect(badRangeEndDefect);
					}
					System.out.println("  ########## Erneute Prfung der korrigierten Container:");
					for(Defect badRangeEndDefect : badRangeEndDefects) {
						checkContainerFile(badRangeEndDefect.getFile());
					}
					final long correctedDefectCount = (badRangeEndDefects.size() - _badRangeEndDefects.size());
					System.out.println("  ########## Anzahl erfolgreich korrigierter Defekte: " + correctedDefectCount);
					_correctedDefectCount += correctedDefectCount;
				}
				if(_badIndexFiles.size() > 0) {
					System.out.println(
							"########## Korrektur von Indexdateien, die fehlerhafte oder unvollstndige Eintrge enthalten. Anzahl: "
							+ _badIndexFiles.size()
					);
					final List<Defect> badIndexFileDefects = _badIndexFiles;
					_badIndexFiles = new ArrayList<Defect>();
					for(Defect badIndexFileDefect : badIndexFileDefects) {
						fixBadIndexFileDefect(badIndexFileDefect);
					}
					System.out.println("  ########## Erneute Prfung der korrigierten Indexdateien:");
					for(Defect badIndexFileDefect : badIndexFileDefects) {
						checkIndexFile(badIndexFileDefect.getFile(), badIndexFileDefect.getContainerIDsInDirectory());
					}
					final long correctedDefectCount = (badIndexFileDefects.size() - _badIndexFiles.size());
					System.out.println("  ########## Anzahl erfolgreich korrigierter Defekte: " + correctedDefectCount);
					_correctedDefectCount += correctedDefectCount;
				}
				System.out.println("########## Gesamtzahl erfolgreich korrigierter Defekte: " + _correctedDefectCount);
				final long t2 = System.currentTimeMillis();
				long fixTime = t2 - t1;
				System.out.println("Dauer der Korrektur:             " + String.format("%.3f Sekunden", fixTime / 1000.0));
			}
			else {
				System.out.println("########## Zur Korrektur der korrigierbaren Defekte bitte das Aufrufargument -reparieren=ja verwenden.");
			}
		}
	}

	private void scan(final File directory) {
		File[] containerFileArray = directory.listFiles(_containerFilenameFilter);
		if(containerFileArray == null) containerFileArray=new File[0];
		List<File> containerFiles = Arrays.asList(containerFileArray);
		if(_checkContainerFiles) {
			for(File containerFile : containerFiles) {
				checkContainerFile(containerFile);
			}
		}
		if(_checkIndexFiles) {
			HashSet<Long> containerIDs = fileListToIdList(containerFiles);
			File[] indexFileArray = directory.listFiles(_indexFilenameFilter);
			if(indexFileArray != null) {
				List<File> indexFiles = Arrays.asList(indexFileArray);
				for(File indexFile : indexFiles) {
					checkIndexFile(indexFile, containerIDs);
				}
			}
		}
		if(_recursiveScan) {
			final File[] subDirectories = directory.listFiles(_directoryFileFilter);
			if(subDirectories != null) {
				for(File subDirectory : subDirectories) {
					//System.out.println("Scanne " + subDirectory);
					scan(subDirectory);
				}
			}
		}
	}

	private HashSet<Long> fileListToIdList(final List<File> containerFiles) {
		final HashSet<Long> hashSet = new HashSet<Long>();
		for(File containerFile : containerFiles) {
			hashSet.add(ContainerFile.getContID(containerFile));
		}
		return hashSet;
	}

	private void checkIndexFile(final File indexFile, final HashSet<Long> containerIDs) {
		_checkedIndexFileCount++;
		final Defect defect = inspectIndexFile(indexFile, containerIDs);
		if(defect == null) {
			_validIndexFileCount++;
		}
		else {
			_defectIndexFileCount++;
			defect.setContainerIDsInDirectory(new HashSet<Long>(containerIDs));
			defect.getCategory().add(defect);
		}
	}

	private Defect inspectIndexFile(final File indexFile, final HashSet<Long> containerIDs) {
		try {
			if(indexFile.canRead()) {
				final RandomAccessFile raf = new RandomAccessFile(indexFile, "r");
				try {
					if(indexFile.getName().startsWith("_didx")) {
						return checkDIDIndexFile(raf, indexFile, containerIDs);
					}
					else if(indexFile.getName().startsWith("_datatime")) {
						return checkDataTimeIndexFile(raf, indexFile, containerIDs);
					}
					else if(indexFile.getName().startsWith("_dataindex")) {
						return checkDataIndexFile(raf, indexFile, containerIDs);
					}
				}
				finally {
					raf.close();
				}
			}
			else {
				return new Defect(indexFile, _cantFix, "Datei nicht lesbar");
			}
		}
		catch(EOFException e) {
			return new Defect(indexFile, _badIndexFiles, "Datei nicht vollstndig");
		}		
		catch(IOException e) {
			return new Defect(indexFile, _cantFix, "Fehler beim Lesen: " + e);
		}
		return new Defect(indexFile, _cantFix, "Unbekannter Dateityp");
	}

	/**
	 * Prft Indexdateien der Klasse {@link de.bsvrz.ars.ars.persistence.index.ATimeMonotoneIndex}. In dieser Klasse ist gefordert, dass die Archivzeit monoton und
	 * die ContainerID streng monoton ueber alle Eintraege hinweg steigend sind.
	 *
	 * @param raf      Geffnete Datei
	 * @param fileName Datei
	 *
	 * @param containerIDs hashSet mit ContainerIDs, die in der IndexDatei vorkommen mssen.
	 * @return Ein Defect, falls die Datei Defekt ist, sonst null
	 *
	 * @throws IOException Eingabe-Ausgabe-Fehler
	 */
	private Defect checkDIDIndexFile(final RandomAccessFile raf, File fileName, final HashSet<Long> containerIDs) throws IOException {
		final long length = raf.length();
		long atmax_old = Long.MIN_VALUE;
		long cid_old = Long.MIN_VALUE;

		final HashSet<Long> containerFileIDs = new HashSet<Long>(containerIDs);

		while(raf.getFilePointer() < length) {
			final long dimin = raf.readLong();
			final long dimax = raf.readLong();
			final long atmin = read6(raf);
			final long atmax = read6(raf);
			final long dtmin = read6(raf);
			final long dtmax = read6(raf);
			final long cid = read5(raf);

			if(!(atmax_old <= atmin)) return new Defect(fileName, _badIndexFiles, "Archivzeit nicht monoton steigend");

			if(!(cid_old < cid)) return new Defect(fileName, _badIndexFiles, "ContainerID nicht streng monoton steigend");

			containerFileIDs.remove(cid);

			atmax_old = atmax;
			cid_old = cid;
		}

		if(containerFileIDs.size() > 0) return new Defect(fileName, _badIndexFiles, "Es kommen nicht alle ContainerIDs in der Indexdatei vor.");

		return null;
	}

	/**
	 * Prft Indexdateien der Klasse {@link de.bsvrz.ars.ars.persistence.index.DataTimeIndex}. In dieser Klasse ist gefordert, dass die Eintrge nach der minimalen
	 * Datenzeit sortiert sind
	 *
	 * @param raf      Geffente Datei
	 * @param fileName Datei
	 *
	 * @param containerIDs hashSet mit ContainerIDs, die in der IndexDatei vorkommen mssen.
	 * @return Ein Defect, falls die Datei Defekt ist, sonst null
	 *
	 * @throws IOException Eingabe-Ausgabe-Fehler
	 */
	private Defect checkDataTimeIndexFile(final RandomAccessFile raf, File fileName, final HashSet<Long> containerIDs) throws IOException {
		final long length = raf.length();
		long dtmin_old = Long.MIN_VALUE;

		final HashSet<Long> indexContainerIds = new HashSet<Long>();
        final HashSet<Long> containerFileIDs = new HashSet<Long>(containerIDs);

		while(raf.getFilePointer() < length) {
			final long dtmin = read6(raf);
			final long dtmax = read6(raf);
			final long dimin = raf.readLong();
			final long dimax = raf.readLong();
			final long cid = read5(raf);

			if(indexContainerIds.contains(cid)) return new Defect(fileName, _badIndexFiles, "ContainerID mehrfach vorhanden");

			if(!(dtmin_old <= dtmin)) return new Defect(fileName, _badIndexFiles, "Datenzeit nicht monoton steigend");

			indexContainerIds.add(cid);
			containerFileIDs.remove(cid);
			dtmin_old = dtmin;
		}
		if(containerFileIDs.size() > 0) return new Defect(fileName, _badIndexFiles, "Es kommen nicht alle ContainerIDs in der Indexdatei vor.");

		return null;
	}

	/**
	 * Prft Indexdateien der Klasse {@link de.bsvrz.ars.ars.persistence.index.DataIndexIndex}. In dieser Klasse ist gefordert, dass die Eintrge nach der
	 * minimalen Datenzeit sortiert sind
	 *
	 * @param raf      Geffente Datei
	 * @param fileName Datei
	 *
     * @param containerIDs hashSet mit ContainerIDs, die in der IndexDatei vorkommen mssen.
	 * @return Ein Defect, falls die Datei Defekt ist, sonst null
	 *
	 * @throws IOException Eingabe-Ausgabe-Fehler
	 */
	private Defect checkDataIndexFile(final RandomAccessFile raf, File fileName, final HashSet<Long> containerIDs) throws IOException {
		final long length = raf.length();
		long dimin_old = Long.MIN_VALUE;

		final HashSet<Long> indexContainerIds = new HashSet<Long>();
        final HashSet<Long> containerFileIDs = new HashSet<Long>(containerIDs);

		while(raf.getFilePointer() < length) {
			final long dimin = raf.readLong();
			final long dimax = raf.readLong();
			final long cid = read5(raf);

			if(indexContainerIds.contains(cid)) return new Defect(fileName, _badIndexFiles, "ContainerID mehrfach vorhanden");

			if(!(dimin_old <= dimin)) return new Defect(fileName, _badIndexFiles, "Datenindex nicht monoton steigend");

			indexContainerIds.add(cid);
			containerFileIDs.remove(cid);
			dimin_old = dimin;
		}
		if(containerFileIDs.size() > 0) return new Defect(fileName, _badIndexFiles, "Es kommen nicht alle ContainerIDs in der Indexdatei vor.");

		return null;
	}

	private static long read5(final RandomAccessFile raf) throws IOException {
		return ((raf.readByte() & 0xffL) << 32) + (raf.readInt() & 0xFFFFFFFFL);
	}

	private static long read6(final RandomAccessFile raf) throws IOException {
		return ((raf.readShort() & 0xffffL) << 32) + (raf.readInt() & 0xFFFFFFFFL);
	}

	private static void fixBadIndexFileDefect(final Defect badIndexFileDefect) {
		try {
			File indexFile = badIndexFileDefect.getFile();
			if(indexFile.getName().startsWith("_didx")) {
				rebuildIndexFile(indexFile, "did");
			}
			else if(indexFile.getName().startsWith("_datatime")) {
				rebuildIndexFile(indexFile, "datatime");
			}
			else if(indexFile.getName().startsWith("_dataindex")) {
				rebuildIndexFile(indexFile, "dataindex");
			}
		}
		catch(Exception e) {
			System.out.println("Fehler beim Wiederherstellen der Indexdatei.");
			e.printStackTrace();
		}
	}

	private static void rebuildIndexFile(File fileName, String indexKind) throws IOException, IndexException, PersistenceException{
		final int REBUILDINDEX_IDX_SIZE = 2048;
		final int INDEX_MEMORY_SIZE = 16 * 1024;
		byte[] tmpBuf = new byte[INDEX_MEMORY_SIZE];

		String path = fileName.getParentFile().getAbsolutePath();

		String[] fNames = new File(path).list(ContainerFile.CONT_FILENAME_FILTER);
		if(fNames != null && fNames.length > 0) {
			Arrays.sort(fNames);

			ATimeMonotoneIndex aTIdx = null;
			DataIndexIndex dIdxIdx = null;
			DataTimeIndex dtidx = null;

			if(indexKind.equals("did")) {
				if(path.endsWith("oa") || path.endsWith("on")) {
					// Bei aktuellen Daten sind Archivzeit und Datenindex stets monoton
					aTIdx = new DIdxATimeMonotoneIndex(REBUILDINDEX_IDX_SIZE, path, DIdxATimeMonotoneIndex.ERASE_OLD);
				}
				else {
					// Bei nachgelieferten Daten kann es Datenindex-Rueckspruenge geben
					aTIdx = new ATimeMonotoneIndex(REBUILDINDEX_IDX_SIZE, path, ATimeMonotoneIndex.ERASE_OLD);
				}
			}
			else if(indexKind.equals("datatime")) {
				dtidx = new DataTimeIndex(REBUILDINDEX_IDX_SIZE, path, tmpBuf, DataTimeIndex.ERASE_OLD);
			}
			else if(indexKind.equals("dataindex")) {
				dIdxIdx = new DataIndexIndex(REBUILDINDEX_IDX_SIZE, path, tmpBuf, DataIndexIndex.ERASE_OLD);
			}
			try {
				for(String contFile : fNames) {
					ContainerFileReader containerFileReader = new ContainerFileReader(new File(path, contFile));
					try {
						long minDidx = containerFileReader.readHeaderAsLong(ContainerHdr.CHP_DATA_IDX_MIN);
						long maxDidx = containerFileReader.readHeaderAsLong(ContainerHdr.CHP_DATA_IDX_MAX);
						long minDtime = containerFileReader.readHeaderAsLong(ContainerHdr.CHP_DATA_TIME_MIN);
						long maxDtime = containerFileReader.readHeaderAsLong(ContainerHdr.CHP_DATA_TIME_MAX);
						long minAtime = containerFileReader.readHeaderAsLong(ContainerHdr.CHP_ARC_TIME_MIN);
						long maxAtime = containerFileReader.readHeaderAsLong(ContainerHdr.CHP_ARC_TIME_MAX);

						boolean rebuild = new Long(ContainerHdr.CHP_DATA_IDX_MIN.getDefaultVal()) == minDidx
						                  || new Long(ContainerHdr.CHP_DATA_IDX_MAX.getDefaultVal()) == maxDidx
						                  || new Long(ContainerHdr.CHP_DATA_TIME_MIN.getDefaultVal()) == minDtime
						                  || new Long(ContainerHdr.CHP_DATA_TIME_MAX.getDefaultVal()) == maxDtime
						                  || new Long(ContainerHdr.CHP_ARC_TIME_MIN.getDefaultVal()) == minAtime
						                  || new Long(ContainerHdr.CHP_ARC_TIME_MAX.getDefaultVal()) == maxAtime;

						if(!rebuild) {
							try {
								if(dtidx != null) dtidx.add(path, tmpBuf, minDtime, maxDtime, minDidx, maxDidx, containerFileReader.getContainerId());
								if(aTIdx != null) {
									aTIdx.add(path, minDidx, maxDidx, minAtime, maxAtime, minDtime, maxDtime, containerFileReader.getContainerId());
								}
								if(dIdxIdx != null) dIdxIdx.add(path, tmpBuf, minDidx, maxDidx, containerFileReader.getContainerId());
							}
							catch(IndexException e) {
								rebuild = true;
							}
						}
						if(rebuild) {
							minDidx = Long.MAX_VALUE;
							maxDidx = Long.MIN_VALUE;
							minAtime = Long.MAX_VALUE;
							maxAtime = Long.MIN_VALUE;
							minDtime = Long.MAX_VALUE;
							maxDtime = Long.MIN_VALUE;
							ContainerDataIterator iter = containerFileReader.getDataIterator();
							while(iter.readNextDataHdr()) {
								long val = iter.getDataIndex();
								minDidx = val < minDidx ? val : minDidx;
								maxDidx = val > maxDidx ? val : maxDidx;
								val = iter.getArcTime();
								minAtime = val < minAtime ? val : minAtime;
								maxAtime = val > maxAtime ? val : maxAtime;
								val = iter.getDataTime();
								minDtime = val < minDtime ? val : minDtime;
								maxDtime = val > maxDtime ? val : maxDtime;
								iter.skipData();
							}
							iter.close();

							if(dtidx != null) dtidx.add(path, tmpBuf, minDtime, maxDtime, minDidx, maxDidx, containerFileReader.getContainerId());
							if(aTIdx != null) aTIdx.add(path, minDidx, maxDidx, minAtime, maxAtime, minDtime, maxDtime, containerFileReader.getContainerId());
							if(dIdxIdx != null) dIdxIdx.add(path, tmpBuf, minDidx, maxDidx, containerFileReader.getContainerId());
						}
					}
					finally {
						containerFileReader.close();
					}
				}
				System.out.println(fileName.toString() + ": Indexdatei wurde repariert");
			}
			finally {
				if(dtidx != null) dtidx.closeIndex(path, tmpBuf);
				if(aTIdx != null) aTIdx.closeIndex(path);
				if(dIdxIdx != null) dIdxIdx.closeIndex(path, tmpBuf);
			}
		}
		else
		{
			System.out.println(fileName + ": Kann Indexdatei nicht reparieren, Containerdateien sind nicht auffindbar.");
		}
	}

	private void checkContainerFile(final File containerFile) {
		//System.out.println("Prfe " + containerFile);
		_checkedContainerFileCount++;
		final Defect defect = inspectContainerFile(containerFile);
		if(defect == null) {
			_validContainerFileCount++;
		}
		else {
			_defectContainerFileCount++;
			defect.getCategory().add(defect);
		}
	}

	private Defect inspectContainerFile(final File containerFile) {
		final BufferedInputStream in;
		final long fileLength = containerFile.length();
		if(fileLength < 384) {
			return new Defect(containerFile, _cantFix, "Unerwartete Dateigre. Erwartet: mindestens 384 Bytes, gefunden: " + fileLength + " Bytes");
		}
		try {
			in = new BufferedInputStream(new FileInputStream(containerFile));
		}
		catch(FileNotFoundException e) {
			return new Defect(containerFile, _cantFix, "Datei kann nicht geffnet werden");
		}
		try {
			byte[] header = new byte[384];
			final int headerLength = in.read(header);
			if(headerLength < 384) {
				return new Defect(
						containerFile, _cantFix, "Unerwartete Anzahl gelesener Headerbytes. Erwartet: 384 Bytes, gefunden: " + headerLength + " Bytes"
				);
			}
			if(header[0] != 'v' || header[1] != '1' || header[2] != '0' || header[3] != '0') {
				return new Defect(
						containerFile,
						_cantFix,
						"Unerwartete Containerversion. Erwartet: 'v100', gefunden: '" + (char)header[0] + (char)header[1] + (char)header[2] + (char)header[3]
						+ "'"
				);
			}
			if(header[4] != 0 || header[5] != 0 || header[6] != 1 || header[7] != 0x78) {
				return new Defect(
						containerFile, _cantFix, "Unerwartete Headerlnge. Erwartet: 0x00000178, gefunden: " + String.format(
						"0x%02x%02x%02x%02x", header[4], header[5], header[6], header[7]
				)
				);
			}
			final char[] deletedParamNameChars = new char[10];
			for(int i = 0; i < deletedParamNameChars.length; i++) {
				deletedParamNameChars[i] = (char)header[0x14d + i];
			}
			final String deletedParamName = String.valueOf(deletedParamNameChars);
			if(!deletedParamName.equals("geloescht=")) {
				return new Defect(containerFile, _cantFix, "Unerwartete Lsch-Kennung. Erwartet: 'geloescht=', gefunden: '" + deletedParamName + "'");
			}
			final boolean containerDeleted;
			if(header[0x14d + 10] == '0') {
				containerDeleted = false;
			}
			else if(header[0x14d + 10] == '1') {
				containerDeleted = true;
			}
			else {
				return new Defect(
						containerFile, _cantFix, "Unerwarteter Wert der Lsch-Kennung. Erwartet: '0' oder '1', gefunden: " + (char)header[0x14d + 10]
				);
			}
			if(containerDeleted) {
				// Beim Lschen eines Containers mit der Methode ContainerFile.deleteContainerData() werden die in der Container-Datei enthaltenen Datenstze
				//  ersetzt durch eine optimierte Aufzhlung der verwendeten Datensatzindizes. Dabei wird die Methode createDataIndexBuffer() verwendet, die
				//  diese optimierte Aufzhlung erzeugt. Diese Methode gibt i.A. ein Byte zuviel zurck. Die deleteContainerData() Methode schreibt die von
				//  createDataIndexBuffer() zurckgelieferten Bytes und verkrzt anschlieend die Datei um grundstzlich ein Byte. In bestimmten Fllen, z.B.,
				//  wenn in der Container-Datei nur ein Datensatz enthalten ist oder nur Datenstze mit Sprngen im Datensatzindex enthalten sind, liefert die
				//  Methode createDataIndexBuffer() allerdings die korrekte Anzahl Bytes zurck, sodass durch das anschlieende Verkrzen der Datei um ein Byte
				//  ein Informationsverlust in der Datei entsteht. Es sollte mglich sein mit Hilfe des maximalen Datensatz-Index, der im Datei-Header
				//  gespeichert ist, das fehlende Byte zu berechnen und die defekte Datei zu reparieren.
				final long mod17remainder = ((fileLength - 384) % 17);
				if(mod17remainder != 0) {
					if(mod17remainder == 16) {
						return new Defect(containerFile, _oneByteShortDefects, "Falsche Dateilnge einer gelschten Datei. Es fehlt ein Byte.");
					}
					else {
						return new Defect(
								containerFile,
								_cantFix,
								"Unerwartete Dateilnge einer gelschten Datei. Erwartet: 384 + 17 * x, gefunden : 384 + 17 * y + " + mod17remainder
						);
					}
				}

				// Beim Lschen eines Containers, in dem als erster Datensatz ein Potentielle-Lcke-Datensatz enthalten ist, wird die optimierte Aufzhlung der
				//  enthaltenen Datensatzindizes mit falschem Endwert (-9223372036854775808) des ersten Bereichs erzeugt. Eine Reparatur der fehlerhaften
				//  Container-Dateien sollte mglich sein, indem der fehlerhafte Endwert -9223372036854775808, durch den entsprechenden Startwert ersetzt wird.
				byte[] range = new byte[17];
				if(in.read(range) == 17) {
					if(range[9] == -128 && range[10] == 0 && range[11] == 0 && range[12] == 0 && range[13] == 0 && range[14] == 0 && range[15] == 0
					   && range[16] == 0) {
						return new Defect(
								containerFile,
								_badRangeEndDefects,
								"Falsche Endangabe eines Datensatzindexbereichs in einem gelschten Container. Erwartet: gltiger Datensatzindex, gelesen: 0x8000000000000000"
						);
					}
				}
			}
		}
		catch(IOException e) {
			e.printStackTrace();
			return new Defect(containerFile, _cantFix, "Unerwartetes Problem: " + e);
		}
		finally {
			try {
				in.close();
			}
			catch(IOException e) {
				// ignored
			}
		}
		return null;
	}

	/**
	 * Beim Lschen eines Containers mit der Methode ContainerFile.deleteContainerData() werden die in der Container-Datei enthaltenen Datenstze ersetzt durch
	 * eine optimierte Aufzhlung der verwendeten Datensatzindizes. Dabei wird die Methode createDataIndexBuffer() verwendet, die diese optimierte Aufzhlung
	 * erzeugt. Diese Methode gibt i.A. ein Byte zuviel zurck. Die deleteContainerData() Methode schreibt die von createDataIndexBuffer() zurckgelieferten Bytes
	 * und verkrzt anschlieend die Datei um grundstzlich ein Byte. In bestimmten Fllen, z.B., wenn in der Container-Datei nur ein Datensatz enthalten ist oder
	 * nur Datenstze mit Sprngen im Datensatzindex enthalten sind, liefert die Methode createDataIndexBuffer() allerdings die korrekte Anzahl Bytes zurck,
	 * sodass durch das anschlieende Verkrzen der Datei um ein Byte ein Informationsverlust in der Datei entsteht. Eine Reparatur der fehlerhaften
	 * Container-Dateien wird mit dieser Methode durchgefhrt, indem mit Hilfe des maximalen Datensatz-Index, der im Datei-Header gespeichert ist, das fehlende
	 * Byte berechnet und die defekte Datei damit korrigiert wird.
	 *
	 * @param defect Identifizierter Defekt einer Containerdatei
	 */
	private void fixOneByteShortDefect(final Defect defect) {
		final RandomAccessFile container;
		try {
			container = new RandomAccessFile(defect.getFile(), "rw");
		}
		catch(FileNotFoundException e) {
			System.out.println(defect.getFile().toString() + ": konnte nicht geffnet werden: " + e);
			return;
		}
		try {
			container.seek(0xA4);
			final byte[] propertyName = new byte[6];
			container.readFully(propertyName);
			final char[] properyValueChars = new char[19];
			for(int i = 0; i < properyValueChars.length; i++) {
				properyValueChars[i] = (char)container.readByte();
			}
			final byte propertySeparator = container.readByte();
			if(propertyName[0] != 'D' || propertyName[1] != 'I' || propertyName[2] != 'm' || propertyName[3] != 'a' || propertyName[4] != 'x'
			   || propertyName[5] != '=') {
				System.out.println(defect.getFile().toString() + ": Reparatur nicht mglich, weil die Kennung DImax= nicht an der erwarteten Stelle ist.");
				return;
			}
			if(propertySeparator != '\n') {
				System.out.println(
						defect.getFile().toString() + ": Reparatur nicht mglich, weil der Wert der Kennung DImax= nicht mit NewLine terminiert ist."
				);
				return;
			}
			final String propertyValueString = String.valueOf(properyValueChars);
			final long propertyValue = Long.parseLong(propertyValueString);
			final long fileLength = container.length();
			container.seek(fileLength - 8);
			final long lastLongValue = container.readLong() << 8;
			if(lastLongValue != Long.MIN_VALUE && lastLongValue != (propertyValue & 0xffffffffffffff00L)) {
				System.out.println(
						defect.getFile().toString() + ": Reparatur nicht mglich, weil die letzten 7 Byte der Datei nicht den erwarteten Werten entsprechen."
				);
				return;
			}
			container.seek(fileLength - 7);
			container.writeLong(propertyValue);
			System.out.println(defect.getFile().toString() + ": Container wurde repariert");
		}
		catch(IOException e) {
			System.out.println(defect.getFile().toString() + ": Unerwarteter Fehler bei der Reparatur: " + e);
			return;
		}
		finally {
			try {
				container.close();
			}
			catch(IOException e) {
				// ignored
			}
		}
	}

	/**
	 * Beim Lschen eines Containers, in dem als erster Datensatz ein Potentielle-Lcke-Datensatz enthalten ist, wird die optimierte Aufzhlung der enthaltenen
	 * Datensatzindizes mit falschem Endwert (-9223372036854775808) des ersten Bereichs erzeugt. Eine Reparatur der fehlerhaften Container-Dateien wird durch diese
	 * Methode durchgefhrt, indem der fehlerhafte Endwert -9223372036854775808, durch den entsprechenden Startwert ersetzt wird.
	 *
	 * @param defect Identifizierter Defekt einer Containerdatei
	 */
	private void fixBadRangeEndDefect(final Defect defect) {
		final RandomAccessFile container;
		try {
			container = new RandomAccessFile(defect.getFile(), "rw");
		}
		catch(FileNotFoundException e) {
			System.out.println(defect.getFile().toString() + ": konnte nicht geffnet werden: " + e);
			return;
		}
		try {
			container.seek(384);
			final long startRangeIndex = container.readLong();
			final byte separator = container.readByte();
			final long badEndRangeIndex = container.readLong();
			if(separator != '-') {
				System.out.println(defect.getFile().toString() + ": Reparatur nicht mglich, weil Separatorzeichen nicht an erwarteter Stelle.");
				return;
			}
			if(badEndRangeIndex != Long.MIN_VALUE) {
				System.out.println(defect.getFile().toString() + ": Reparatur nicht mglich, weil Bereichsende des ersten Bereichs nicht defekt ist.");
				return;
			}
			container.seek(384 + 9);
			container.writeLong(startRangeIndex);
			System.out.println(defect.getFile().toString() + ": Container wurde repariert");
		}
		catch(IOException e) {
			System.out.println(defect.getFile().toString() + ": Unerwarteter Fehler bei der Reparatur: " + e);
			return;
		}
		finally {
			try {
				container.close();
			}
			catch(IOException e) {
				// ignored
			}
		}
	}

	private static class Defect {

		private final File _file;

		private final List<Defect> _category;

		private final String _message;

		private HashSet<Long> _containerIDsInDirectory;

		public Defect(final File file, final List<Defect> category, final String message) {
			_file = file;
			_category = category;
			_message = message;
			System.out.println(file.toString() + ": " + message);
		}


		public List<Defect> getCategory() {
			return _category;
		}

		public File getFile() {
			return _file;
		}

		public String getMessage() {
			return _message;
		}

		public HashSet<Long> getContainerIDsInDirectory() {
			return _containerIDsInDirectory;
		}

		public void setContainerIDsInDirectory(final HashSet<Long> containerIDsInDirectory) {
			_containerIDsInDirectory = containerIDsInDirectory;
		}
	}
}
